from __future__ import division

fornac_template = """
<!DOCTYPE html>
<meta charset="utf-8">

This is an RNA container.
<div id='rna_ss'> </div>
This after the RNA container.

    <link rel='stylesheet' type='text/css' href='https://raw.githubusercontent.com/pkerpedjiev/fornac/master/css/fornac.css' />
    <script type='text/javascript' src='https://code.jquery.com/jquery-2.1.4.min.js'></script>
    <script type='text/javascript' src='https://cdnjs.cloudflare.com/ajax/libs/d3/3.5.6/d3.min.js'></script>
    <script type='text/javascript' src='https://rawgit.com/pkerpedjiev/fornac/master/js/fornac.js'></script>

    <script type='text/javascript'>
        var container = new fornac.FornaContainer("#rna_ss",
            {{'applyForce': true, 'allowPanningAndZooming': true, "initialSize": [500,800], 'cssFileLocation': "https://raw.githubusercontent.com/pkerpedjiev/fornac/master/css/fornac.css"}});

        var options = {{'structure': '{}',
                        'sequence': '{}'}};

        colorStr = "{}" 

        container.addRNA(options.structure, options);
        cs = ColorScheme(colorStr);

        container.addCustomColors(cs.colorsJson);
        container.changeColorScheme('custom'); 
    </script>
"""


def create_fornac_page_for_structure(bg, color_string):
    """
    Create a fornac page dispalying this structure. The dotbracket string
    and sequence will automatically be extracted from the BulgeGraph.

    Colors can be specified as a dictionary containing floating 
    point values. These will be uniformly scaled according to the color
    scale passed in.

    :param color_string: The color string to be passed to fornac.
                          e.g. "11-12: red 14-17: rgb(14,15,120)"
    :return: The html text of the resulting web page.
    """
    return fornac_template.format(bg.to_dotbracket_string(),
                                  bg.seq, color_string)


def scale_colors(colors_dict, cmap=None, reverse=False):
    '''
    A dictionary with values containing scalars which need to 
    be scaled according to some color map. The keys are irrelevant.

    The color map will be normalized to the range of values within 
    color_dict.

    :param colors_dict: The dictionary containing the values to be color scaled.
    :param cmap: A color map to be used to scale the colors.
    :param reverse: Reverse the color map
    :return: Another dictionary containing rgb triples as values.
    '''

    if cmap is None:
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap('Blues')

    values = colors_dict.values()
    min_value = min(values)
    max_value = max(values)

    new_dict = {}
    for key in colors_dict:
        if reverse:
            color = cmap(
                1 - ((colors_dict[key] - min_value) / (max_value - min_value)))
        else:
            color = cmap(
                (colors_dict[key] - min_value) / (max_value - min_value))
        new_dict[key] = (int(255 * color[0]),
                         int(255 * color[1]), int(255 * color[2]))

    return new_dict


def element_to_nucleotide_colors(bg, element_colors):
    '''
    Convert a dictionary of per-element colors to a dictionary of per-nucleotide colors

    :param element_colors: A dictionary of element colors (e.g. {'i0': (255,0,0), 'm1': {255,255,255)}
    :return: A dictionary of nucleotide numbers: (e.g {1: (14,15,120), 2: (255,255,255)})
    '''
    new_dict = {}
    for key in element_colors:
        for res in bg.define_residue_num_iterator(key):
            new_dict[res] = element_colors[key]

    return new_dict


def nucleotide_colors_to_fornac_color_string(nucleotide_colors):
    '''
    Convert a dictionary of per nucleotide colors to a fornac
    color string.

    :param nucleotide_colors: A dictionary with nucleotide numbers as keys and colors as values.
                              (e.g. {1: (255,0,0), 2: (255,255,0)})
    :return: A color string (e.g "1:rgb(255,0,0) 2:rgb(255,0,0)")
    '''
    color_string = ""
    for key in nucleotide_colors:
        color_string += "{}:rgb({},{},{}) ".format(key, nucleotide_colors[key][0],
                                                   nucleotide_colors[key][1],
                                                   nucleotide_colors[key][2])

    return color_string



"""
Take a BulgeGraph and return a copy of it with cofold splitpoints inserted.
"""
import logging

from logging_exceptions import log_to_exception


from ._graph_construction import remove_vertex, relabel_node
from ..utilities.exceptions import GraphConstructionError


log = logging.getLogger(__name__)


# This module belongs to the BulgeGraph creation machinenery and is allowed to access its private members
# pylint: disable=protected-access


def split_at_cofold_cutpoints(bg, cutpoints):
    """
    Multiple sequences should not be connected along the backbone.

    We have constructed the bulge graph, as if they were connected along the backbone, so
    now we have to split it.
    """

    for splitpoint in cutpoints:
        element_left = bg.get_node_from_residue_num(splitpoint)
        element_right = bg.get_node_from_residue_num(splitpoint + 1)
        if element_left[0] in "ft" or element_right[0] in "ft":
            if element_left[0] == "t" and element_left[0] != "t":
                continue  # Splitpoint already implemented
            elif element_right[0] == "f" and element_left[0] != "f":
                continue  # Splitpoint already implemented
            else:
                # No cofold structure. First sequence is disconnected from rest
                e = GraphConstructionError("Cannot create BulgeGraph. Found two sequences not "
                                           "connected by any base-pair.")
                with log_to_exception(log, e):
                    log.error("Trying to split between %s and %s",
                              element_left, element_right)
                raise e
            return
        elif element_left[0] == "i" or element_right[0] == "i":
            _split_interior_loop(bg, splitpoint, element_left, element_right)
        elif element_left != element_right:
            _split_between_elements(
                        bg, splitpoint, element_left, element_right)
        elif element_left[0] == "s":
            _split_inside_stem(bg, splitpoint, element_left)
        else:
            _split_inside_loop(bg, splitpoint, element_left)
        bg._node_to_resnum = {}
    if not _is_connected(bg):
        raise GraphConstructionError("Cannot create BulgeGraph. Found two sequences not connected by any "
                                     " base-pair.")


def _is_connected(bg):
    if not bg.defines:
        return True  # We define an empty Graph as connected.
    start_node = list(bg.defines.keys())[0]
    known_nodes = set([start_node])
    pending = list(bg.edges[start_node])
    while pending:
        next_node = pending.pop()
        if next_node in known_nodes:
            continue
        pending.extend(bg.edges[next_node])
        known_nodes.add(next_node)
    log.info("Testing connectivity: connected component =?= all nodes:\n%s =?= %s",
             list(sorted(known_nodes)), list(sorted(set(bg.defines.keys()))))
    return known_nodes == set(bg.defines.keys())


def _split_between_elements(bg, splitpoint, element_left, element_right):
    log.debug("Before splitting between %s and %s at %s: %s", element_left,
              element_right, splitpoint, bg.defines)
    if element_left[0] in "mh":
        next3 = _next_available_element_name(bg, "t")
        relabel_node(bg, element_left, next3)
        if element_left[0] != "h":
            _remove_edge(bg, next3, element_right)
    elif element_right[0] in "mh":
        next5 = _next_available_element_name(bg, "f")
        relabel_node(bg, element_right, next5)
        if element_right[0] != "h":
            _remove_edge(bg, next5, element_left)
    else:
        assert element_left[0] == "s" and element_right[0] == "s"
        # Zero-length i or m element!
        connections = bg.edges[element_left] & bg.edges[element_right]
        # If multiple ml connections exist, find the one which is 0-len
        # Remove the highest numbered element first
        # We know that adjacent stems are connected
        # by either an i-loop or a 0-len ml.
        for connection in reversed(sorted(connections)):
            if connection[0] == "i":
                break
            if not bg.defines[connection]:
                assert splitpoint in bg.defines[element_left]
                break
        else:
            raise GraphConstructionError(
                        "Cannot split at cofold cutpoint. "
                        "No (suitable) connection between "
                        "{} and {}.".format(element_left, element_right))
        # pylint: disable=undefined-loop-variable
        if connection[0] == "m":
            # Just remove it without replacement
            remove_vertex(bg, connection)
        else:
            assert connection[0] == "i"
            # Replace i by ml (this is then located on the other strand than the splitpoint)
            next_ml = _next_available_element_name(bg, "m")
            assert next_ml not in bg.defines
            relabel_node(bg, connection, next_ml)


def _split_inside_loop(bg, splitpoint, element):
    if element[0] in "hm":
        from_, to_ = bg.defines[element]
        stem_left = bg.get_node_from_residue_num(from_ - 1)
        stem_right = bg.get_node_from_residue_num(to_ + 1)

        next3 = _next_available_element_name(bg, "t")
        next5 = _next_available_element_name(bg, "f")
        bg.defines[next3] = [from_, splitpoint]
        bg.defines[next5] = [splitpoint + 1, to_]
        _add_edge(bg, stem_left, next3)
        _add_edge(bg, next5, stem_right)
        remove_vertex(bg, element)
    else:
        assert False


def _split_inside_stem(bg, splitpoint, element):
    assert element[0] == "s"
    log.debug("Split inside stem %s at %s", element, splitpoint)
    if splitpoint == bg.defines[element][1]:
        # Nothing needs to be done. 2 strands split at end
        log.debug("Nothing to do")
        return
    if splitpoint < bg.defines[element][1]:
        # Splitpoint in forward strand:
        define1 = [bg.defines[element][0], splitpoint,
                   bg.pairing_partner(splitpoint), bg.defines[element][3]]
        define2 = [splitpoint + 1, bg.defines[element][1],
                   bg.defines[element][2], bg.pairing_partner(splitpoint + 1)]
        log.debug("Split in forward strand")
    else:
        # Splitpoint in backwards strand:
        define1 = [bg.defines[element][0], bg.pairing_partner(splitpoint + 1),
                   splitpoint + 1, bg.defines[element][3]]
        define2 = [bg.pairing_partner(splitpoint), bg.defines[element][1],
                   bg.defines[element][2], splitpoint]
        log.debug("Split in backwards strand")
    edges1 = []
    edges2 = []

    for edge in bg.edges[element]:
        log.debug("Checking edge %s with define %s connected to %s",
                  edge, bg.defines[edge], bg.edges[edge])
        flank = bg.flanking_nucleotides(edge)
        found = 0
        if define1[0] in flank or define1[3] in flank:
            edges1.append(edge)
            found += 1
        if define2[1] in flank or define2[2] in flank:
            edges2.append(edge)
            found += 1
        if found != 1:
            log.error("For stem %s with define %s and cutpoint %s:",
                      element, bg.defines[element], splitpoint)
            log.error("Edge %s, with flanking nts %s, define1 %s, "
                      "define2 %s", edge, bg.flanking_nucleotides(edge),
                      define1, define2)
            assert False
    remove_vertex(bg, element)
    next_s1 = _next_available_element_name(bg, "s")
    bg.defines[next_s1] = define1
    next_m = _next_available_element_name(bg, "m")
    bg.defines[next_m] = []
    next_s2 = _next_available_element_name(bg, "s")
    bg.defines[next_s2] = define2

    for e1 in edges1:
        bg.edges[e1].add(next_s1)
    for e2 in edges2:
        bg.edges[e2].add(next_s2)
    edges1.append(next_m)
    edges2.append(next_m)
    bg.edges[next_s1] = set(edges1)
    bg.edges[next_s2] = set(edges2)
    bg.edges[next_m] = set([next_s1, next_s2])


def _next_available_element_name(bg, element_type):
    """
    :param element_type: A single letter ("t", "f", "s"...)
    """
    i = 0
    while True:
        name = "{}{}".format(element_type, i)
        if name not in bg.defines:
            return name
        i += 1


def _remove_edge(bg, from_element, to_element):
    bg.edges[from_element].remove(to_element)
    bg.edges[to_element].remove(from_element)


def _add_edge(bg, from_element, to_element):
    bg.edges[from_element].add(to_element)
    bg.edges[to_element].add(from_element)


def _split_interior_loop_at_side(bg, splitpoint, strand, other_strand, stems):
    """
    Called by _split_at_cofold_cutpoints
    """
    next_ml = _next_available_element_name(bg, "m")
    next_a = _next_available_element_name(bg, "t")
    next_b = _next_available_element_name(bg, "f")

    if other_strand[0] > other_strand[1]:
        bg.defines[next_ml] = []
    else:
        bg.defines[next_ml] = other_strand
    _add_edge(bg, next_ml, stems[0])
    _add_edge(bg, next_ml, stems[1])

    if splitpoint >= strand[0]:
        bg.defines[next_a] = [strand[0], splitpoint]
        _add_edge(bg, next_a, stems[0])
    if splitpoint < strand[1]:
        bg.defines[next_b] = [splitpoint + 1, strand[1]]
        _add_edge(bg, next_b, stems[1])


def _split_interior_loop(bg, splitpoint, element_left, element_right):
    if element_left[0] == "i":
        iloop = element_left
    elif element_right[0] == "i":
        iloop = element_right
    else:
        assert False
    c = bg.connections(iloop)
    s1 = bg.defines[c[0]]
    s2 = bg.defines[c[1]]
    forward_strand = [s1[1] + 1, s2[0] - 1]
    back_strand = [s2[3] + 1, s1[2] - 1]
    if forward_strand[0] - 1 <= splitpoint <= forward_strand[1]:
        # Split forward strand, relabel backwards strand to multiloop.
        _split_interior_loop_at_side(bg, splitpoint,
                                     forward_strand, back_strand, c)
    elif back_strand[0] - 1 <= splitpoint <= back_strand[1]:
        _split_interior_loop_at_side(bg, splitpoint,
                                     back_strand, forward_strand, [c[1], c[0]])
    else:
        assert False
    remove_vertex(bg, iloop)



"""
Code for reading and writing the forgi configuration files.

To configure forgi, run forgi_config.py
"""


import json
import os
import os.path
import logging
import appdirs

log = logging.getLogger(__name__)

dirs = appdirs.AppDirs("forgi", "TBI")  # pylint: disable=C0103
ALLOWED_KEY_VALUES = {"PDB_ANNOTATION_TOOL": ["MC-Annotate", "DSSR", "forgi"]}


def iter_configfiles():
    """
    Iterate over config file locations, from lowest priority to highest priority.
    The file does not have to exist.
    """
    for directory in [dirs.site_config_dir, dirs.user_config_dir]:
        filename = os.path.join(directory, "config.json")
        yield filename


def read_config():
    """
    Read all configuration files and return a configuration dictionary

    :returns: A dictionary.
    """
    config = {}
    for filename in iter_configfiles():
        try:
            with open(filename) as f:
                conf = json.load(f)
        except (OSError, IOError):
            log.debug("No configuration file present at %s", filename)
        else:
            log.debug("Reading configuration from %s", filename)
            config.update(conf)
    return config



"""
A module where all costum exceptions should be defines.

We put exceptions into this seperate module, so they can be
imported everywhere without risking circular imports
"""


class GraphConstructionError(ValueError):
    """
    Exceptions raised if the BulgeGraph could be constructed from the given input.
    """
    pass


class GraphIntegrityError(ValueError):
    """
    Exception raised if a BulgeGraph was found to be in an inconsistent or faulty state.
    """
    pass


class CgConstructionError(GraphConstructionError):
    """
    Exceptions raised if the CoarseGrainRNA could be constructed from the given input.

    Raised for Errors related to the 3D information.
    """
    pass


class CgIntegrityError(GraphIntegrityError):
    """
    Exception raised if a BulgeGraphCoarseGrainRNA was found to be in an inconsistent or faulty state,
    related to the 3D Information
    """
    pass



#!usr/bin/env python
from __future__ import division

import logging
import argparse
import copy
import csv
from collections import defaultdict
import pandas

import forgi.graph.bulge_graph as fgb
import forgi.utilities.commandline_utils as fuc
import forgi.threedee.utilities.vector as ftuv
from forgi.utilities.exceptions import GraphIntegrityError
from forgi.utilities.exceptions import GraphConstructionError
from forgi.utilities.numbered_dotbracket import NumberedDotbracket

log = logging.getLogger(__name__)

"""
Classify all pseudoknots found in the input 3D structure based on their shapes
according to Reidys et al. concept into H-type, kissing hairpin,...

Additionally, calculate the angles between stems in the poseudoknots.
Annotation Tools: MC-mcannotate or DSSR (enable stacking analysis)

See:
Reidys, C.M. et al., 2011. Topology and prediction of RNA pseudoknots.
Bioinformatics, 27(8), pp.1076-1085.

Input:
python pseudoknot_analyzer.py filename.cg --minlength #

:param minlength: Minimum length of each stem, default=2
"""

PK_CHECK = ["[","]","{","}","<",">"]
CLOSE_TO_OPEN = {
        ")":"(",
        "]":"[",
        "}":"{",
        ">":"<"
        }

def identification_pseudoknot(rna, without_pk, with_pk, unfold, minlength):
    """
    Return isolated pseudoknots from the dotbracket string

    :param rna: A forgi CoarseGrainRNA object
    :param without_pk: Counter for structures without a pseudoknots
    :param with_pk: Counter for structures with a pseudoknots
    :param unfold: Counter for unfolded Structures
    """
    if isinstance(rna, fgb.BulgeGraph):
        residue_numbers = copy.copy(rna.seq._seqids)
        dotbracket = NumberedDotbracket(rna.to_dotbracket_string(), residue_numbers)
    else:
        dotbracket = rna

    #variables
    modified_dotbracket = NumberedDotbracket()
    pseudoknots_dictionary = {}
    counter_pseudoknots = 0
    pseudoknots = []
    pseudoknot = NumberedDotbracket()
    #general, round, square, curve, arrow
    pseudoknot_open = [0,0,0,0,0]
    pseudoknot_check = False

    #remove all dots:
    dotbracket = dotbracket.without_unpaired()
    log.info("Bracket-structure: {}".format(dotbracket))

    if not dotbracket:
        unfold += 1
        return  without_pk, with_pk, unfold, pseudoknots
    else:
        pass

    dotbracket = dotbracket.without_short_helices(minlength)
    dotbracket = dotbracket.without_unpaired()

    if not dotbracket:
        unfold += 1
        return  without_pk, with_pk, unfold, pseudoknots
    else:
        pass

    #remove consecutive basepairs = "condense"
    dotbracket  = dotbracket.condensed()
    log.info("Condensed-bracket-structure: {}".format(dotbracket))

    if set(dotbracket)-set("()[]{}<>"):
        raise ValueError("Pseudoknots with more than 4 types of "
                         "brackets are not supported.")

    #discard every structure without pseudoknot
    if not any (pk in dotbracket for pk in PK_CHECK):
        without_pk += 1
        print("Without Psedudoknot: {}".format(without_pk))
    #for structures with pseudoknots
    else:
        #for number, residue in residue_numbers.items():
        with_pk += 1
        for bracket in dotbracket:
            log.info("actual bracket {}".format(bracket))
            #CLOSE BRACKETS
            if bracket in CLOSE_TO_OPEN:
                log.info("CLOSE pk-check {}".format(pseudoknot_check))
                #close pseudoknot
                if pseudoknot_check == True:
                    openbracket = CLOSE_TO_OPEN[bracket]
                    log.debug("bracket:{}; pseudoknot[-1]:{}; openbracket: {} "
                    .format(bracket, pseudoknot[-1],openbracket))

                    #update open brackets
                    pseudoknot_open[0] -= 1
                    if bracket == PK_CHECK[1]:
                        pseudoknot_open[2] -= 1
                    elif bracket == PK_CHECK[3]:
                        pseudoknot_open[3] -= 1
                    elif bracket == PK_CHECK[5]:
                        pseudoknot_open[4]-= 1
                    else:
                        pseudoknot_open[1]-= 1

                    #pseudoknot uptdate
                    if openbracket == pseudoknot[-1]:
                        pseudoknot = pseudoknot[:-1]
                        log.info("check: {}; pseudoknot yet: {}; open: {}"
                        .format(pseudoknot_check, pseudoknot, pseudoknot_open))
                    else:
                        pseudoknot = pseudoknot + bracket
                        #check if the pseudoknot is finished
                        if not any(pseudoknot_open):
                            pseudoknots.append(pseudoknot)
                            counter_pseudoknots += 1
                            pseudoknots_dictionary.update({counter_pseudoknots:pseudoknot})
                            pseudoknot_check = False
                            pseudoknot = NumberedDotbracket()
                            pseudoknot_open = [0,0,0,0,0]
                            log.info("FINISHED PSEUDOKNOT: check: {}; pseudoknot yet: {}; pseudoknots: {}"
                            .format(pseudoknot_check, pseudoknot, pseudoknots))
                        #check if bracket from modified_dotbracket is needed
                        elif pseudoknot_open[0]==0 or any(v < 0 for v in pseudoknot_open):
                            pseudoknot = modified_dotbracket[-1] + pseudoknot
                            modified_dotbracket = modified_dotbracket[:-1]
                            pseudoknot_open[0] += 1
                            pseudoknot_open[1] += 1
                            log.info("modified_dotbracket:{}; pseudoknot yet: {}; open: {} "
                            .format(modified_dotbracket, pseudoknot, pseudoknot_open))
                        #continue
                        else:
                            log.info("check: {}; pseudoknot yet: {}; open: {}"
                            .format(pseudoknot_check, pseudoknot, pseudoknot_open))

                #normal close
                elif modified_dotbracket[-1] == CLOSE_TO_OPEN[bracket]:
                    modified_dotbracket = modified_dotbracket[:-1]
                    log.info("skip last modified_dotbracket {}".format(modified_dotbracket))
                else:
                    log.info("Something strange happen - close function")
                    assert False

            #OPEN BRACKETS
            #if a pesoudoknot starts [ (can only start with "[")
            elif (bracket == PK_CHECK[0]) and pseudoknot_check == False:
                #position before can only be a (, all ) should be already skipped
                pseudoknot = modified_dotbracket[-1] + bracket
                modified_dotbracket = modified_dotbracket[:-1]
                pseudoknot_check = True
                pseudoknot_open[0] += 2 #([
                pseudoknot_open[1] += 1 #(
                pseudoknot_open[2] += 1 #[
                log.info("PK-START: modified_dotbracket: {}; check: {}; pseudoknot jet: {}, open {}"
                            .format(modified_dotbracket, pseudoknot_check, pseudoknot, pseudoknot_open))
            else:
                #if PK is True
                if pseudoknot_check == True:
                    pseudoknot += bracket
                    pseudoknot_open[0] += 1
                    if bracket == PK_CHECK[0]:
                        pseudoknot_open[2] += 1
                    elif bracket == PK_CHECK[2]:
                        pseudoknot_open[3] += 1
                    elif bracket == PK_CHECK[4]:
                        pseudoknot_open[4] += 1
                    else:
                        pseudoknot_open[1] += 1
                    log.info("check: {}; pseudoknot yet: {}, open: {}"
                    .format(pseudoknot_check, pseudoknot, pseudoknot_open))
                #if PK is False
                else:
                    modified_dotbracket +=bracket
                    log.info("open: {}; append modified_dotbracke: {}"
                    .format(bracket,modified_dotbracket))

    if not modified_dotbracket:
        log.info("modified_dotbracket is empty - WORKS CORRECT HERE")
    else:
        log.info("Something went wrong due skipping () pairs, rest: {}"
        .format(modified_dotbracket))
        assert False

    return without_pk, with_pk, unfold, pseudoknots


def classify_pseudoknot_genus1(pseudoknot):
    """
    Return the classes of a pseudoknot with genus1

    :param pseudoknot: Current genus1-"other" pseudoknot
    """
    if pseudoknot == "([)]":
        return "g1_H_type"
    elif pseudoknot == "([)(])":
        return "g1_Kissinghairpin"
    elif pseudoknot == "([{)]}":
        return "g1_L"
    elif pseudoknot == "([{)(]})":
        return "g1_M"
    else:
        log.debug("Other pseudoknot: %s", str(pseudoknot))
        return "other"

def update_pk_classes(classes1, classes2):
    """
    Update the count of the current pseudoknot class

    :param classes1: Dictionary with pseudoknot classes genus1
    :param classes2: Dictionary with pseudoknot classes genus2
    """
    for k,v in classes2.items():
        classes1[k]+=v

def classify_pseudoknots(pseudoknots):
    """
    Return the class of a pseudoknot with and counts them

    :param pseudoknots: A list of pseudoknots
    """
    pk_classes = {
        "g1_H_type":[],
        "g1_Kissinghairpin":[],
        "g1_L":[],
        "g1_M":[],
        "g2_H_type":[],
        "g2_Kissinghairpin":[],
        "g2_L":[],
        "g2_M":[],
    }
    other = []
    for pseudoknot in pseudoknots:
        log.debug("Classifying %s for genus 1", pseudoknot)
        pk_type = classify_pseudoknot_genus1(pseudoknot)
        #if the pseudoknot is no other - append it
        if pk_type != "other":
            pk_classes[pk_type].append(pseudoknot)
        #if the pseudoknot is "other" clean it again
        else:
            log.debug("Other %s", pseudoknot)
            _,_,_,cleaned_pks = identification_pseudoknot(pseudoknot,0,0,0,0)
            #if cleand version is different classify again
            if cleaned_pks != [pseudoknot]:
                classes_cleaned, cleaned_other = classify_pseudoknots(cleaned_pks)
                other+=cleaned_other
                update_pk_classes(pk_classes, classes_cleaned)
            #otherwise same with genus 2
            else:
                cleaned_pk, = cleaned_pks #Almost same as cleaned_pk = cleaned_pks
                classes_g2, remaining_pks = classify_pseudoknot_genus2(cleaned_pk)
                other+=remaining_pks
                update_pk_classes(pk_classes, classes_g2)
    return pk_classes, other

def classify_pseudoknot_genus2(pseudoknot):
    """
    Return the classes of a pseudoknot with genus2 and count the occurences

    :param pseudoknot: Current genus1-"other" pseudoknot
    """
    log.debug("Classifying %s for genus 2", pseudoknot)
    pk_classes = {
        "g2_H_type":[],
        "g2_Kissinghairpin":[],
        "g2_L":[],
        "g2_M":[],
    }
    while True:
        old_pseudoknot = pseudoknot
        if "([)]" in pseudoknot:
            occurrences, pseudoknot = pseudoknot.without_substr("([)]")
            pk_classes["g2_H_type"]+=occurrences
        if "([)(])" in pseudoknot:
            occurrences, pseudoknot = pseudoknot.without_substr("([)(])")
            pk_classes["g2_Kissinghairpin"]+=occurrences
        if "([{)]}" in pseudoknot:
            occurrences, pseudoknot = pseudoknot.without_substr("([{)]}")
            pk_classes["g2_L"]+=occurrences
        if pseudoknot == "([{)(]})":
            occurrences, pseudoknot = pseudoknot.without_substr("([{)(]})")
            pk_classes["g2_M"]+=occurrences
        if pseudoknot == old_pseudoknot:
            break
    if pseudoknot:
        _,_,_,cleaned_pks = identification_pseudoknot(pseudoknot,0,0,0,0)
        remaining_pks = []
        if cleaned_pks != [pseudoknot]:
            for pk in cleaned_pks:
                pk_classes_2, remaining = classify_pseudoknot_genus2(pk)
                update_pk_classes(pk_classes, pk_classes_2)
                remaining_pks+=remaining
        else:
            remaining_pks = [pseudoknot]
    else:
        remaining_pks = []
    return pk_classes, remaining_pks


def stem_parameters(stem, rna, side):
    """
    :param stem: E.g. 's1'
    :param rna: A forgi CoarseGrainRNA object
    :param side: 0 or 1. See explaination above.
    """
    position = rna.coords[stem][side]
    direction = rna.coords[stem][not side]-position
    return position, direction

def stem_after_next_ml(rna, pos, before):
    ml_found = False
    for elem in rna.iter_elements_along_backbone(pos):
        if elem[0]=="m":
            ml_found = True
        elif ml_found:
            assert elem[0]=="s"
            return elem
        elif elem == before:
            return None
    return None

def extend_pk_description(dataset, filename, pk_type, rna, pk, pk_number):
    """
    Return a extended descripiton of current pseudoknot in the current files
    e.g. angles between stems

    :param dataset:  Current dataset that will be updated
    :param filename: Filename of the current structure
    :parma pk_type: Class of the pseudoknot
    :param rna: A forgi CoarseGrainRNA object
    :param pk: Structure of the pseudoknot, a NumberedDotbracket object,
               in a condensed (shadow-like) representation.
               This representation always contains the most 5' basepair.
    :param pk_number: consecutive number of the pseudoknot
    """
    domains = rna.get_domains()
    helices = domains["rods"] # A list of elements, e.g. ["s0", "i0", "s1"]
    log.debug("Helices: %s", helices)
    #rna.log(logging.WARNING)
    stems_5p = []
    stems_3p = []

    nums = []
    log.debug("pk Residue numbers %s", pk.residue_numbers)
    log.debug("pk helix ends %s", pk.helix_ends)

    for i, resnum in enumerate(pk.residue_numbers):
        num = rna.seq.to_integer(resnum)
        nums.append(num)
        element_5p = rna.get_node_from_residue_num(num)
        stems_5p.append(element_5p)

        num2 = rna.seq.to_integer(pk.helix_ends[i])
        log.debug("num %s nums2 %s", num, num2)
        element_3p =rna.get_node_from_residue_num(num2)
        stems_3p.append(element_3p)
    log.debug("nums %s", nums)
    for i, stem1_5p in enumerate(stems_5p):
        dataset["Filename"].append(filename)
        dataset["rnaname"] = rna.name
        dataset["pk_type"].append(pk_type)
        dataset["pk_id"].append(pk_number)
        dataset["angle_nr"].append(i)
        if pk_type == "other":
            dataset["pk_structure"].append(str(pk))
        else:
            dataset["pk_structure"].append("")
        #is this the first occurrence of stem in stems?
        if stems_5p.index(stem1_5p)==i:
            #first occurrence. Strand 0, look at 3' end of helix
            stem1 = stems_3p[i]
            strand = 0
        else:
            assert i>stems_5p.index(stem1_5p)
            stem1 = stem1_5p
            strand = 1
        try:
            stem2_5p = stems_5p[i+1]
        except IndexError:
            stem2_5p = stems_5p[0]
            outside_pk = True
        else:
            outside_pk = False
        if outside_pk or stems_5p.index(stem2_5p)==i+1:
            #first occurrence
            stem2 = stem2_5p
            strand2 = 0
        else:
            strand2 = 1
            if outside_pk:
                stem2 = stems_3p[0]
            else:
                stem2 = stems_3p[i+1]
        log.debug("Stem 5' %s, 3' %s, stem1 %s stem2 %s", stems_5p, stems_3p, stem1, stem2)
        # enable stacking analysis via DSSR
        # differentiate between stacking (True), no stacking (False) and brakes
        # within/aorund the pseudoknot (-1) incl. 'virtual' angles e.g. H-Type angle_type3
        ml_stack=[]
        if rna.dssr:
            nc_bps = list(rna.dssr.noncanonical_pairs())
            nc_dict = defaultdict(list)
            for nt1, nt2, typ in nc_bps:
                nc_dict[nt1].append((nt2, typ))
                nc_dict[nt2].append((nt1, typ))
            stacking_loops = rna.dssr.stacking_loops()
            start_found = 0
            connection = []
            stacking = None
            branch = None
            log.debug("Checking %s and %s for stacking, strand %s", stem1, stem2, strand)
            for elem in rna.iter_elements_along_backbone(): #walk along the backbone
                if start_found == strand+1:
                    if branch:
                        log.debug("in branch: elem %s, branch %s, stacking %s", elem, branch, stacking)
                        if elem == branch:
                            log.debug("End branch at %s", elem)
                            branch = None
                            log.debug("Branch end")
                        continue
                    if elem[0] != "s":
                        connection.append(elem)
                        if rna.defines[elem] and rna.defines[elem][-1] in rna.backbone_breaks_after:
                            stacking = -1
                        if elem not in stacking_loops and stacking != -1:
                            stacking = False
                    elif elem == stem2:
                        if stacking is None:
                            stacking = True
                        log.debug("Found second stem, elem %s, stacking %s", elem, stacking)
                        break
                    elif elem[0] == "s" and connection:
                        branch = elem
                        if rna.defines[elem][-1] in rna.backbone_breaks_after:
                            stacking = -1
                    log.debug("elem %s, stacking %s, branch %s", elem, stacking, branch)
                elif elem == stem1:
                    start_found += 1
                    if rna.defines[elem][strand*2+1] in rna.backbone_breaks_after:
                        stacking = -1
                    log.debug("First stem, elem %s, stacking %s", elem, stacking)
            else:
                log.debug("End iteration, stacking->-1")
                stacking = -1
            log.debug("Finally, stacking = %s", stacking)
            # more detailed stacking (including backbone brackes within and around the pseudoknot)
            dataset["this_loop_stacking_dssr"].append(stacking)
            dataset["connecting_loops"].append(",".join(connection))

            # more genereal stacking information
            connecting_loops = rna.edges[stem1]&rna.edges[stem2]
            for loop in connecting_loops:
                if loop in stacking_loops:
                    ml_stack.append(loop)
            stacks = rna.dssr.coaxial_stacks()
            log.info("Stacks: %s", stacks)
            for stack in stacks:
                if stem1 in stack and stem2 in stack:
                    # the two stems stack, but we do not specify along which
                    # multiloop segment they stack.
                    dataset["is_stacking_dssr"].append(True)
                    break
            else:
                dataset["is_stacking_dssr"].append(False)

            # Does the connection form base-triples with the stem?
            stem1_triples=0
            stem2_triples=0
            aminors1 = 0
            aminors2 = 0
            aminors = list(rna.dssr.aminor_interactions())
            for elem in connection:
                for nt in rna.define_residue_num_iterator(elem,seq_ids=True):
                    if (nt, stem1) in aminors:
                        aminors1+=1
                        log.debug("AMinor %s (%s), %s", nt, elem, stem1)
                    elif (nt, stem2) in aminors:
                        aminors2+=1
                        log.debug("AMinor %s (%s), %s", nt, elem, stem2)
                    else:
                        for partner, typ in nc_dict[nt]:
                            if rna.get_elem(partner)==stem1:
                                log.debug("base_triple %s, %s: %s-%s (%s)", elem, stem1, nt,partner,typ)
                                stem1_triples+=1
                            elif rna.get_elem(partner)==stem2:
                                log.debug("base_triple %s, %s: %s-%s (%s)", elem, stem2, nt,partner,typ)
                                stem2_triples+=1
            log.debug("%s has a length of %s and %s triples", stem1, rna.stem_length(stem1),stem1_triples)
            log.debug("%s has a length of %s and %s triples", stem2, rna.stem_length(stem2),stem2_triples)
            dataset["stem1_basetripleperc_dssr"].append(stem1_triples/rna.stem_length(stem1))
            dataset["stem2_basetripleperc_dssr"].append(stem2_triples/rna.stem_length(stem2))
            dataset["stem1_aminorperc_dssr"].append(aminors1/rna.stem_length(stem1))
            dataset["stem2_aminorperc_dssr"].append(aminors2/rna.stem_length(stem2))

        else:
            dataset["is_stacking_dssr"].append(float("nan"))
            dataset["this_loop_stacking_dssr"].append(float("nan"))
            dataset["connecting_loops"].append("")
            dataset["stem1_basetripleperc_dssr"].append(float("nan"))
            dataset["stem2_basetripleperc_dssr"].append(float("nan"))
            dataset["stem1_aminorperc_dssr"].append(float("nan"))
            dataset["stem2_aminorperc_dssr"].append(float("nan"))

        dataset["stacking_loops"].append(",".join(ml_stack))

        pos1, dir1 = stem_parameters(stem1, rna, not strand)
        pos2, dir2 = stem_parameters(stem2, rna, strand2)
        dataset["stem1"].append(stem1)
        dataset["stem2"].append(stem2)

        dataset["angle_between_stems"].append(ftuv.vec_angle(dir1, dir2))
        dataset["distance_between"].append(ftuv.vec_distance(pos1, pos2))

        next_stem = None
        if not outside_pk:
            next_stem = stem_after_next_ml(rna, nums[i], before=stem2)
            if next_stem==stem2:
                next_stem  = None
        if next_stem:
            posN, dirN = stem_parameters(next_stem, rna, 0)
            dataset["angle_to_next"].append(ftuv.vec_angle(dir1, dirN))
            dataset["distance_to_next"].append(ftuv.vec_distance(pos1, posN))
            dataset["next_stem"].append(next_stem)
        else:
            dataset["angle_to_next"].append("")
            dataset["distance_to_next"].append("")
            dataset["next_stem"].append("")
        dataset["outside_pk"].append(outside_pk)



def main():
    parser = fuc.get_rna_input_parser("Find pseudoknots in RNA structures, "
                                      "classify them into shapes and analyze "
                                      "their 3D architecturre.", "+",
                                      parser_kwargs={"conflict_handler":"resolve"})
    parser.add_argument("--pseudoknots", action="store_true", help=argparse.SUPPRESS)
    parser.add_argument("--outfile-mode", choices=["w","a"], default='w', help="Overwrite ('w') or append ('a') to output file")
    parser.add_argument("--minlength", type= int, help= "Minimum length of each stem. "
                            "Stems with fewer base-pairs are treated as unpaired.",
                            default = 2)

    args =  parser.parse_args()
    args.pseudoknots=True
    rnas, filenames = fuc.cgs_from_args(args, rna_type="any",return_filenames = True, skip_errors=True)


    #variables for statistics
    unfold = 0
    without_pk = 0
    with_pk = 0
    pseudoknot_dataset = []
    pseudoknot_dataset_extended = defaultdict(list)

    for pos, rna in enumerate(rnas):
        try:
            without_pk, with_pk, unfold, pseudoknots = identification_pseudoknot\
                                            (rna, without_pk, with_pk, unfold,args.minlength)

            #count types of pseudoknots
            total_pk_g1 = len(pseudoknots)

            pk_sortclasses = {}
            other = 0
            other_pk =  []

            pk_classes, other = classify_pseudoknots(pseudoknots)

            print("pk_classes:")
            for key, pks in sorted(pk_classes.items()):
                print("{:<17s} {}\t{}".format(key, len(pks), ", ".join(map(str, pks))))
            print("other:"+"\t"+", ".join(map(str,other)))

            filename = str(filenames[pos]).split("/")[-1]

            entry={}
            for key, pks in pk_classes.items():
                entry[key] = len(pks)
            entry["other"] = len(other)
            entry["PK_other_structures"] = ",".join(map(str, other))
            entry["filename"] = filename
            entry["rnaname"] = rna.name
            print(filename)
            pseudoknot_dataset.append(entry)
            pk_id = 0
            for key, pks in pk_classes.items():
                for pk in pks:
                    pk_id+=1
                    extend_pk_description(pseudoknot_dataset_extended, filename,
                                          key, rna, pk, pk_id)


            for pk in other:
                pk_id+=1
                extend_pk_description(pseudoknot_dataset_extended, filename,
                                      "other", rna, pk, pk_id)
        except GraphIntegrityError:
            log.exception("Ignoring RNA %s; GraphIntegrityError", rna.name)
        except GraphConstructionError:
            log.exception("Ignoring RNA %s; GraphConstructionError", rna.name)
        except Exception:
            log.error("Error processing %s", rna.name)
            raise
    df1 = pandas.DataFrame(pseudoknot_dataset)
    df1.to_csv("pseudoknot_identification_genus2.csv", mode=args.outfile_mode, header=args.outfile_mode!="a", sep="\t")

    df2 = pandas.DataFrame(pseudoknot_dataset_extended)
    df2.to_csv("pseudoknot_identification_extended_genus2.csv", mode=args.outfile_mode, header=args.outfile_mode!="a", sep="\t")


    print("Structures with Pseudoknots: {}".format(with_pk))
    print("Structures without Pseudoknots: {}".format(without_pk))
    print("Structures unfold: {}".format(unfold))

if __name__ == "__main__":
    main()



"""
The package forgi.projection is a subpackage of forgi responsible for creating 2D projections of 
an RNA 3D structure. 

The class .projection2d.Projection2D is initialized from a 
forgi.threedee.model.coarse_grain.CoarseGrainRNA object. 
It contains methods for plotting and analyzing the projection.

The module forgi.projection.hausdorff consists of two parts.
On the one hand there are very general functions for calculating the Hausdorff distance
between two images represented as boolean matrices. On the other hand there are heuristic 
functions built on top of the before mentioned functions, responsible for calculating the optimal 
alignment of a CoarseGrainRNA and an image, so that the Hausdorff distance between the projection
of the RNA and the image is minimal.
"""



from __future__ import print_function

import sys
import argparse
import logging
import os.path
import contextlib
import warnings
import textwrap
import copy

import numpy as np

from .exceptions import GraphConstructionError
from logging_exceptions import log_to_exception
import logging_exceptions


import forgi.graph.bulge_graph as fgb
import forgi.graph.sequence as fgs
import forgi.threedee.model.coarse_grain as ftmc

log = logging.getLogger(__name__)


class WrongFileFormat(ValueError):
    """
    Error raised if the input file has the wrong file format or the file format could not be detected.
    """


def get_rna_input_parser(helptext, nargs=1, rna_type="any", enable_logging=True, parser_kwargs={}):
    parser = argparse.ArgumentParser(description=helptext, **parser_kwargs)
    if nargs == 1:
        helptext = "One file containing an RNA.\n"
    elif isinstance(nargs, int) and nargs > 1:
        helptext = "{:d} files containing one RNA each.\n".format(nargs)
    elif nargs == "+":
        helptext = "One or more files containing one or more RNAs each.\n"
    elif nargs == "*":
        helptext = "Zero or more files containing one or more RNAs each.\n"    
    else:
        raise ValueError(
            "get_parser_any_cgsg does not support nargs={}".format(nargs))

    if rna_type != "only_cg":
        fileformats = ["pdb files"]
    else:
        fileformats = []
    if rna_type != "pdb":
        fileformats.append("forgi cg files")
        if rna_type != "3d" and rna_type != "only_cg":
            fileformats.append("forgi bg files")
            fileformats.append("fasta files")
    if rna_type == "any":
        fileformats.append("dotbracketfiles")
    n = len(fileformats)
    helptext += "\n".join(textwrap.wrap(
        "Supported Filetypes are: {}".format(", ".join(fileformats)), 55))
    if rna_type == "any":
        helptext += ("Alternatively you can supply a dotbracket-string\n "
                     "(containing only the characters '.()[]{}&') from the commandline.\n")
    parser.add_argument("rna", nargs=nargs, type=str, help=helptext)
    parser.add_argument("--keep-length-one-stems", action="store_true",
                        help="For all input formats except forgi bg/cg files,\n"
                             "this controlls whether stems of length one are \n"
                             "dissolved to unpaired regions (default) \n"
                             "or kept (if this option is present).\n"
                             "In  the case of input in forgi-format,\n"
                             "the RNA from the file is not modified.")
    if rna_type != "only_cg":
        pdb_input_group = parser.add_argument_group("Options for loading of PDB files",
                                                    description="These options only take effect, "
                                                    "if the input RNA is in pdb file format.")
        pdb_input_group.add_argument("--pseudoknots", action="store_true",
                                     help="Allow pseudoknots when extracting the structure\nfrom PDB files.")
        pdb_input_group.add_argument("--chains", type=str,
                                     help="When reading pdb-files: Only extract the given chain(s). Comma-seperated")
        pdb_input_group.add_argument("--pdb-secondary-structure", type=str, default="",
                                     help="When reading a single chain from a pdb-files: \nEnforce the secondary structure given as dotbracket\n string. (This only works, if --chain is given!)")
        pdb_input_group.add_argument("--pdb-annotation-tool", type=str, default=None,
                                     help="What program to use for detecting basepairs in PDB/ MMCIF structures."
                                     " This commandline option overrides the value in the config file (if present)."
                                     "If this is not present and no config-file is given, we try to detect the installed programs.")
        pdb_input_group.add_argument("--pdb-allow-www-query", action="store_true",
                                     help="Usually, if modified residues/ ligand with an unknown 3-letter code are encountered in PDB files, "
                                          "they are removed from the chain and a log message of severity INFO is issued. "
                                          "With this option, we first try to query the PDBeChem database for the 3-letter code, to "
                                          "see whether or not it is a modified residue that can be converted to its standard parent "
                                          " and should be part of the chain.")

    if enable_logging:
        verbosity_group = parser.add_argument_group(
            "Control verbosity of logging output")
        logging_exceptions.update_parser(verbosity_group)
    return parser


def cgs_from_args(args, rna_type="any", enable_logging=True,
                  return_filenames=False, skip_errors=False):
    """
    Given an Namespace from argparse, return a list of CoarseGrainRNA objects
    (or BulgeGraph objects).

    :param args: A argparse Namespace.
    :param rna_type: See documentation of load_rna
    :param enable_logging: In addition to loading the CoarseGrainRNA objects,
                    call logging.basicConfig() and use args.verbose
                    and args.debug
                    to set the level for different loggers.
    :param return_filenames: Instead of returning a list of rnas,
                    return a tuple `rnas, filenames`,
                    where filenames is a list of filenames of equal length as
                    the returned list of CoarseGrainRNA objects and with
                    corresponding order.
                    If a file contains two seperate RNA molecules, the
                    filename will thus be found twice in the list of filenames.
    :param skip_errors: Boolean. Log GraphConstructionErrors and continue with
                    the next filename instead of letting the error propagate.

    Usage::

        parser = get_rna_input_parser()
        args = parser.parse_args()
        rnas = cgs_from_args(args)

    """
    cgs, fns = [], []
    for value in iter_cgs_from_args(args, rna_type, enable_logging,
                                    return_filenames, skip_errors):
        if return_filenames:
            cgs.append(value[0])
            fns.append(value[1])
        else:
            cgs.append(value)
    if return_filenames:
        return cgs, fns
    return cgs


def iter_cgs_from_args(args, rna_type="any", enable_logging=True,
                       return_filenames=False, skip_errors=False):
    """
    Given an Namespace from argparse, iterate over the corresponding of CoarseGrainRNA objects
    (or BulgeGraph objects).

    :param args: A argparse Namespace.
    :param rna_type: See documentation of load_rna
    :param enable_logging: In addition to loading the CoarseGrainRNA objects,
                    call logging.basicConfig() and use args.verbose
                    and args.debug
                    to set the level for different loggers.
    :param return_filenames: Instead of yielding just the rnas,
                    yield tuples `rna, filename`,
                    If a file contains two seperate RNA molecules, the same
                    filename will be yeilded twice.
    :param skip_errors: Boolean. Log GraphConstructionErrors and continue with
                    the next filename instead of letting the error propagate.

    Usage::

        parser = get_rna_input_parser()
        args = parser.parse_args()
        rnas = cgs_from_args(args)

    """
    if enable_logging:
        logging.basicConfig(
            format="%(levelname)s:%(name)s.%(funcName)s[%(lineno)d]: %(message)s")
        logging_exceptions.config_from_args(args)
    for rna in args.rna:
        log.debug("Load RNA %s", rna)
        if rna_type == "only_cg":
            args.chains = None
            args.pseudoknots = None
            args.pdb_secondary_structure = None
            args.pdb_annotation_tool = None
            args.pdb_allow_www_query = False
        if args.chains:
            load_chains = args.chains.split(",")
        else:
            load_chains = None
        try:
            cg_or_cgs = load_rna(rna, rna_type=rna_type, allow_many=True,
                                 pdb_chain=load_chains,
                                 pdb_remove_pk=not args.pseudoknots, pdb_dotbracket=args.pdb_secondary_structure,
                                 dissolve_length_one_stems=not args.keep_length_one_stems,
                                 pdb_annotation_tool=args.pdb_annotation_tool,
                                 pdb_allow_www_query=args.pdb_allow_www_query)
        except GraphConstructionError:
            if not skip_errors:
                log.error("An error occurred while loading the file %s", rna)
                raise
            else:
                log.exception("The PDB %s was skipped due to the following error", rna)
        else:
            for cg in cg_or_cgs:
                if return_filenames:
                    yield cg, rna
                else:
                    yield cg


def sniff_filetype(file):
    line = next(file)
    # PDB
    if line.startswith("ATOM") or line.startswith("HEADER") or line.startswith("HETATM") or line.startswith("REMARK"):
        return "pdb"
    line = line.strip()
    # We allow comments and empty lines in all files except PDB files.
    while not line or line.startswith("#"):
        line = next(file).strip()
    if line.startswith("name"):
        return "forgi"
    elif line.startswith("_") or line == "loop_":
        return "cif"
    elif line.startswith(">"):
        return "fasta"
    elif line[0].isdigit():
        return "bpseq"
    else:
        # Is it a bp-seq or mmcif file with header?
        try:
            while True:
                line = next(file).strip()
                if line.startswith("#"):
                    continue
                if line.startswith("_"):
                    return "cif"
                if line[0].isdigit():
                    try:
                        d1, nt, d2 = line.split()
                        if d1.isdigit() and nt in "AUGCTIaugcti" and d2.isdigit():
                            return "bpseq"
                    except (TypeError, ValueError):
                        pass
                    break
        except StopIteration:
            pass
        return "other"


def load_rna(filename, rna_type="any", allow_many=True, pdb_chain=None,
             pdb_remove_pk=True, pdb_dotbracket="",
             dissolve_length_one_stems=True,
             pdb_annotation_tool=None, pdb_allow_www_query=False):
    """
    :param rna_type: One of "any", and "3d" and "pdb"

                     *  "any": Return either BulgeGraph or CoarseGrainRNA object,
                               depending on the input format
                     *  "only_cg": Only accept cg-files.
                     *  "3d":  Return CoarseGrainRNA objects,
                               if the file contains 3D information,
                               raise an error otherwise
                     *   "pdb": only accept pdb files

    :param allow_many: If True, return a list. If False, return a single
                       CoarseGrainRNA object or raise a WrongFileFormat,
                       if more than one RNA is present.
    :param pdb_chain: Extract the given chain from the file.
                      Only applicable if filename corresponds to a pdb file
    :param pdb_remove_pk: Detect pseudoknot-free structures from the pdb.
    :param pdb_dotbracket: Only applicable, if filename corresponds to a pdb file and pdb_chain is given.
    :param dissolve_length_one_stems: Ignored if input is in forgi bg/cg format.
    :param pdb_annotation_tool: Use DSSR, MC-Annotate or forgi heuristic for
                    basepair-detection in PDB/MMCIF files (None for auto-detect).
                    Ignored for other file-types.

    :retuns: A list of RNAs or a single RNA
    """
    # Is filename a dotbracket string and not a filename?
    if all(c in ".()[]{}&" for c in filename):
        # A dotbracket-string was provided via the commandline
        if not rna_type == "any":
            warnings.warn("Cannot treat '{}' as dotbracket string, since we need a sequence. "
                          "Trying to treat it as a filename instead...".format(filename))
        else:
            log.info(
                "Assuming RNA %s is a dotbracketstring and not a file.", filename)
            bg = fgb.BulgeGraph.from_dotbracket(
                filename, dissolve_length_one_stems=dissolve_length_one_stems)
            if allow_many:
                return [bg]
            else:
                return bg
    with open(filename) as rnafile:
        filetype = sniff_filetype(rnafile)
    if rna_type == "pdb" and filetype not in ["pdb", "cif"]:
        raise WrongFileFormat(
            "Only PDB files (*.pdb/.cif) are accepted, but file {} has type {}.".format(filename, filetype))
    if rna_type == "only_cg" and filetype != "forgi":
        raise WrongFileFormat(
            "Only forgi cg files are accepted, but file {} has type {}.".format(filename, filetype))
    if filetype == "forgi":
        cg = ftmc.CoarseGrainRNA.from_bg_file(filename)
        if rna_type in ["3d", "only_cg"] and not cg.coords.is_filled: # pylint: disable=E1101
            raise WrongFileFormat(
                "File {} does not contain all 3D coordinates!".format(filename))
        if allow_many:
            return [cg]
        else:
            return cg
    elif filetype == "pdb" or filetype == "cif":
        if pdb_chain:
            cgs = ftmc.CoarseGrainRNA.from_pdb(filename, load_chains=pdb_chain,
                                               remove_pseudoknots=pdb_remove_pk and not pdb_dotbracket,
                                               secondary_structure=pdb_dotbracket,
                                               dissolve_length_one_stems=dissolve_length_one_stems,
                                               filetype=filetype, annotation_tool=pdb_annotation_tool,
                                               query_PDBeChem=pdb_allow_www_query)
        else:
            if pdb_dotbracket:
                raise ValueError(
                    "pdb_dotbracket requires a chain to be given to avoid ambiguity.")
            cgs = ftmc.CoarseGrainRNA.from_pdb(filename, remove_pseudoknots=pdb_remove_pk,
                                               dissolve_length_one_stems=dissolve_length_one_stems,
                                               filetype=filetype, annotation_tool=pdb_annotation_tool,
                                               query_PDBeChem=pdb_allow_www_query)
        if allow_many:
            return cgs
        else:
            if len(cgs) > 1:
                raise WrongFileFormat("More than one connected RNA component in pdb file {}: {}".format(
                    filename, [cg.name for cg in cgs]))
            return cgs[0]
    # elif filetype=="mmcif":
    #    raise WrongFileFormat("MMCIF files are not yet supported.")
    elif filetype == "bpseq":
        if rna_type == "3d":
            raise WrongFileFormat(
                "bpseq file {} is not supported. We need 3D coordinates!".format(filename))
        with open(filename, 'r') as f:
            text = f.read()
            try:
                int(text[0])
            except ValueError:
                i = text.find("\n1 ")
                text = text[i + 1:]
        bg = ftmc.CoarseGrainRNA.from_bpseq_str(
            text, dissolve_length_one_stems=dissolve_length_one_stems)
        if allow_many:
            return [bg]
        else:
            return bg
    elif filetype == "fasta" or filetype == "other":
        if rna_type == "3d":
            raise WrongFileFormat(
                "Fasta(like) file {} is not supported. We need 3D coordinates!".format(filename))
        try:
            bgs = ftmc.CoarseGrainRNA.from_fasta(
                filename, dissolve_length_one_stems=dissolve_length_one_stems)
        except Exception as e:
            with log_to_exception(log, e):
                log.critical("Could not parse file %r.", filename)
                if filetype == "other":
                    log.critical(
                        "We assumed file %r to be some fasta-variant or dotbracket file, but an error occurred during parsing.", filename)
            raise
        if allow_many:
            return bgs
        else:
            if len(bgs) > 1:
                raise WrongFileFormat(
                    "More than one RNA found in fasta/ dotbracket file {}.".format(filename))
            return bgs[0]



@contextlib.contextmanager
def open_for_out(filename=None, force=False):
    "From http://stackoverflow.com/a/17603000/5069869"
    if filename and filename != '-':
        if not force and os.path.isfile(filename):
            raise IOError(
                "Cannot create file {}. File exists.".format(filename))
        fh = open(filename, 'w')
    else:
        fh = sys.stdout
    try:
        yield fh
    finally:
        if fh is not sys.stdout:
            fh.close()


@contextlib.contextmanager
def hide_traceback(error_class=WrongFileFormat):
    """
    A context manager that catches errors of the specified class and outputs
    only the error message before calling sys.exit(1).
    """
    try:
        yield
    except error_class as e:
        # Integration with logging_exceptions
        if hasattr(e, "log"):
            for record in e.log:
                logging.getLogger(record.name).handle(record)
        print("Error of type {} occurred. Aborting.".format(type(e).__name__))
        print(e, file=sys.stderr)
        sys.exit(1)

def db_to_constraint(dbstri):
    """
    Convert a dotbracket string with dashes representing missing residues and
    dots representing unpaired nucleotides to a constraint that can be used
    with RNAfold.

    This means that unpaired nucleotides are forced unpaired and missing
    residues are free to form basepairs.
    """
    dbstri = dbstri.replace(".", "x").replace("-",".")
    return ''.join( "x" if c not in "()." else c for c in dbstri )

def insert_pk_into_stru(refolded, with_pk):
    """
    Replaces '.' characters in refolded with pkseudoknots from with_pk where possible.

    :param with_pk: A dotbracket string, possibly with '-' characters
    :param refolded: The same dbstring as with_pk, but without pseudoknots and with '-' replaced by one of '().
    """
    out=[]
    for i,c in enumerate(refolded):
        if c != with_pk[i]:
            if c==".":
                if with_pk[i]=="-":
                    out.append(c)
                else:
                    out.append(with_pk[i])
            else:
                assert with_pk[i]=="-"
                out.append(c)
        else:
            out.append(c)
    return "".join(out)

def with_missing_refolded(cg):
    """
    This requires the ViennaRNA package.

    Uses RNAfold
    """
    domains = cg.get_domains()
    fasta_lines = cg.to_fasta_string(include_missing=True).splitlines()
    assert len(fasta_lines)==3
    constraint = db_to_constraint(fasta_lines[2])
    try:
        import RNA
    except ImportError as e:
        raise ImportError(str(e)+"\nPlease install the ViennaRNA package to fold missing residues.")
    fc = RNA.fold_compound(fasta_lines[1], RNA.md())
    fc.hc_add_from_db(constraint)
    stru, energy = fc.mfe()

    stru=insert_pk_into_stru(stru, fasta_lines[2])
    log.info("Energy of constraint secondary structyure is %s", energy)
    new_fasta="\n".join([fasta_lines[0], fasta_lines[1], stru])
    new_cg, = ftmc.CoarseGrainRNA.from_fasta_text(new_fasta)
    new_cg._seq = fgs.missing_to_normal(cg.seq)
    # TODO: Copy coordinates, but take potentially changed 2D structure into account.

    #new_cg.sampled = copy.deepcopy(cg.sampled)
    #new_cg.longrange = copy.deepcopy(cg.longrange)
    #new_cg.interacting_residues = copy.deepcopy(cg.interacting_residues)
    #new_cg.vposs = {k:copy.deepcopy(v) for k,v in cg.vposs.items() if k[0]!="s"}

    return new_cg



from __future__ import print_function
from __future__ import division
from builtins import zip
from past.builtins import basestring

import Bio
import forgi.graph.residue as fgr
import forgi.threedee.model.coarse_grain as ftmc
import forgi.threedee.utilities.pdb as ftup
import forgi.threedee.utilities.vector as ftuv
import math
import matplotlib.pyplot as plt
import numpy as np
import logging
import itertools
import colorsys

log = logging.getLogger(__name__)

def circles(x, y, s, c='b', ax=None, vmin=None, vmax=None,labels=[], **kwargs):
    """
    Make a scatter of circles plot of x vs y, where x and y are sequence
    like objects of the same lengths. The size of circles are in data scale.

    :param x,y: scalar or array_like, shape (n, )
                Input data
    :param s: scalar or array_like, shape (n, )
              Radius of circle in data scale (ie. in data unit)
    :param c: color or sequence of color, optional, default : 'b'
              `c` can be a single color format string, or a sequence of color
              specifications of length `N`, or a sequence of `N` numbers to be
              mapped to colors using the `cmap` and `norm` specified via kwargs.
              Note that `c` should not be a single numeric RGB or
              RGBA sequence because that is indistinguishable from an array of
              values to be colormapped.  `c` can be a 2-D array in which the
              rows are RGB or RGBA, however.
    :param ax: Axes object, optional, default: None
               Parent axes of the plot. It uses gca() if not specified.
    :param vmin, vmax: scalar, optional, default: None
        `vmin` and `vmax` are used in conjunction with `norm` to normalize
        luminance data.  If either are `None`, the min and max of the
        color array is used.  (Note if you pass a `norm` instance, your
        settings for `vmin` and `vmax` will be ignored.)
    :param kwargs: `~matplotlib.collections.Collection` properties
        eg. alpha, edgecolors, facecolors, linewidths, linestyles, norm, cmap

    :returns: paths : `~matplotlib.collections.PathCollection`


    Examples ::

        a = np.arange(11)
        circles(a, a, a*0.2, c=a, alpha=0.5, edgecolor='none')

    This code by ??? is under [The BSD 3-Clause License]
    (http://opensource.org/licenses/BSD-3-Clause)

    """
    from matplotlib.patches import Circle
    from matplotlib.collections import PatchCollection
    #import matplotlib.colors as colors

    if ax is None:
        ax = plt.gca()

    if isinstance(c, basestring):
        color = c     # ie. use colors.colorConverter.to_rgba_array(c)
    else:
        color = None  # use cmap, norm after collection is created
    kwargs.update(color=color)

    if np.isscalar(x):
        patches = [Circle((x, y), s), ]
    elif np.isscalar(s):
        patches = [Circle((x_, y_), s) for x_, y_ in zip(x, y)]
    else:
        patches = [Circle((x_, y_), s_) for x_, y_, s_ in zip(x, y, s)]

    collection = PatchCollection(patches, **kwargs)

    if color is None:
        collection.set_array(np.asarray(c))
        if vmin is not None or vmax is not None:
            collection.set_clim(vmin, vmax)

    ax.add_collection(collection)
    ax.autoscale_view()
    return collection

def _clashfree_annot_pos(pos, coords):
    for c in coords:
        dist = ftuv.vec_distance(c, pos)
        #log.debug("vec_dist=%s", dist)
        if dist<14:
            return False
    return True

def _annotate_rna_plot(ax, cg, coords, annotations, text_kwargs):
    # Plot annotations
    annot_dict = { elem:elem for elem in cg.defines}
    if annotations is None:
        annot_dict={elem:"" for elem in cg.defines}
    else:
        annot_dict.update(annotations)
    stem_coords = {}
    for stem in cg.stem_iterator():
        stem_start =  np.mean([coords[cg.defines[stem][0]-1],
                               coords[cg.defines[stem][3]-1]],
                               axis=0)
        stem_end =  np.mean([coords[cg.defines[stem][1]-1],
                               coords[cg.defines[stem][2]-1]],
                               axis=0)
        stem_center = np.mean([stem_start, stem_end], axis=0)
        stem_coords[stem]=(stem_start, stem_center, stem_end)
        if annot_dict[stem]:
            stem_vec = stem_end - stem_start
            norm_vec = (stem_vec[1], -stem_vec[0])
            norm_vec/=ftuv.magnitude(norm_vec)
            annot_pos = np.array(stem_center)+23*norm_vec
            #log.debug("Checking clashfree for %s, %s", stem, annot_pos)
            if not _clashfree_annot_pos(annot_pos, coords):
                log.debug("Cannot annotate %s as %s ON THE RIGHT HAND SIDE, because of insufficient space. Trying left side...", stem, annot_dict[stem])
                annot_pos = np.array(stem_center)-23*norm_vec
                #log.debug("Checking clashfree OTHER SIDE for %s, %s", stem, annot_pos)
                if not _clashfree_annot_pos(annot_pos, coords):
                    log.info("Cannot annotate %s as '%s', because of insufficient space.", stem, annot_dict[stem])
                    annot_pos = None
            #log.debug("%s", annot_pos)
            if annot_pos is not None:
                ax.annotate(annot_dict[stem], xy=annot_pos,
                            ha="center", va="center", **text_kwargs )
    for hloop in cg.hloop_iterator():
        hc = []
        for nt in cg.define_residue_num_iterator(hloop, adjacent=True):
            hc.append(coords[nt-1])
        annot_pos = np.mean(hc, axis=0)
        if _clashfree_annot_pos(annot_pos, coords):
            ax.annotate(annot_dict[hloop], xy=annot_pos,
                        ha="center", va="center", **text_kwargs )
        else:
            log.info("Cannot annotate %s as '%s' ON THE INSIDE, because of insufficient space. Trying outside...", hloop, annot_dict[hloop])
            nt1, nt2 = cg.define_a(hloop)
            start = np.mean([coords[nt1-1], coords[nt2-1]], axis=0)
            vec = annot_pos-start
            annot_pos = annot_pos+vec*3
            if _clashfree_annot_pos(annot_pos, coords):
                ax.annotate(annot_dict[hloop], xy=annot_pos,
                            ha="center", va="center", **text_kwargs )
            else:
                log.info("Cannot annotate %s as '%s', because of insufficient space.", hloop, annot_dict[hloop])
    for iloop in cg.iloop_iterator():
        s1, s2 = cg.connections(iloop)
        annot_pos = np.mean([ stem_coords[s1][2], stem_coords[s2][0]], axis=0)
        if _clashfree_annot_pos(annot_pos, coords):
            ax.annotate(annot_dict[iloop], xy=annot_pos,
                        ha="center", va="center", **text_kwargs )
        else:
            log.debug("Cannot annotate %s as '%s' ON THE INSIDE, because of insufficient space. Trying outside...", iloop, annot_dict[iloop])
            loop_vec = stem_coords[s2][0] - stem_coords[s1][2]
            norm_vec = (loop_vec[1], -loop_vec[0])
            norm_vec/=ftuv.magnitude(norm_vec)
            annot_pos_p = np.array(annot_pos)+25*norm_vec
            annot_pos_m = np.array(annot_pos)-25*norm_vec
            # iloops can be asymmetric (more nts on one strand.)
            # plot the label on the strand with more nts.
            plus=0
            minus=0
            for nt in cg.define_residue_num_iterator(iloop):
                if ftuv.vec_distance(annot_pos_p, coords[nt-1])<ftuv.vec_distance(annot_pos_m, coords[nt-1]):
                    plus+=1
                else:
                    minus+=1
            if plus>minus:
                if _clashfree_annot_pos(annot_pos_p, coords):
                    ax.annotate(annot_dict[iloop], xy=annot_pos_p,
                                ha="center", va="center", **text_kwargs )
                else:
                    log.info("Cannot annotate %s as '%s' (only trying inside and right side), because of insufficient space.", iloop, annot_dict[iloop])

            else:
                if _clashfree_annot_pos(annot_pos_m, coords):
                    ax.annotate(annot_dict[iloop], xy=annot_pos_m,
                                ha="center", va="center", **text_kwargs )
                else:
                    log.info("Cannot annotate %s as '%s' (only trying inside and left side), because of insufficient space.", iloop, annot_dict[iloop])
    for mloop in itertools.chain(cg.floop_iterator(), cg.tloop_iterator(), cg.mloop_iterator()):
        nt1, nt2 = cg.define_a(mloop)
        res = list(cg.define_residue_num_iterator(mloop))
        if len(res)==0:
            anchor = np.mean([coords[nt1-1], coords[nt2-1]], axis=0)
        elif len(res)%2==1:
            anchor = coords[res[int(len(res)//2)]-1]
        else:
            anchor =  np.mean([ coords[res[int(len(res)//2)-1]-1],
                                coords[res[int(len(res)//2)]-1] ],
                              axis=0)
        loop_vec = coords[nt1-1] - coords[nt2-1]
        norm_vec = (loop_vec[1], -loop_vec[0])
        norm_vec/=ftuv.magnitude(norm_vec)
        annot_pos = anchor - norm_vec*18
        if _clashfree_annot_pos(annot_pos, coords):
            ax.annotate(annot_dict[mloop], xy=annot_pos,
                        ha="center", va="center", **text_kwargs )
        else:
            log.info("Cannot annotate %s as '%s' , because of insufficient space.", mloop, annot_dict[mloop])

def plot_rna(cg, ax=None, offset=(0, 0), text_kwargs={}, backbone_kwargs={},
             basepair_kwargs={}, color=True, lighten=0, annotations={}):
    '''
    Plot an RNA structure given a set of nucleotide coordinates

    .. note::

        This function calls set_axis_off on the axis. You can revert this by
        using ax.set_axis_on() if you like to see the axis.

    :param cg: A forgi.threedee.model.coarse_grain.CoarseGrainRNA structure
    :param ax: A matplotlib plotting area
    :param offset: Offset the plot by these coordinates. If a simple True is passed in, then
                   offset by the current width of the plot
    :param text_kwargs: keyword arguments passed to matplotlib.pyplot.annotate
                        for plotting of the sequence
    :param backbone_kwargs: keyword arguments passed to matplotlib.pyplot.plot
                        for plotting of the backbone links
    :param basepair_kwargs: keyword arguments passed to matplotlib.pyplot.plot
                        for plotting of the basepair links
    :param lighten: Make circles lighter. A percent value where 1 makes
                    everything white and 0 leaves the colors unchanged
    :param annotations: A dictionary {elem_name: string} or None.
                        By default, the element names (e.g. "s0") are plotted
                        next to the element. This dictionary can be used to
                        override the default element names by costum strings.
                        To remove individual annotations, assign an empty string to the key.
                        To remove all annotations, set this to None.

                        .. warning::

                            Annotations are not shown, if there is not enough space.
                            Annotations not shown are logged with level INFO
    :return: (ax, coords) The axes and the coordinates for each nucleotide
    '''
    log.info("Starting to plot RNA...")
    import RNA
    import matplotlib.colors as mc
    RNA.cvar.rna_plot_type = 1

    coords = []
    #colors = []
    #circles = []

    bp_string = cg.to_dotbracket_string()
    # get the type of element of each nucleotide
    el_string = cg.to_element_string()
    # i.e. eeesssshhhhsssseeee
    el_to_color = {'f': 'orange',
                   't': 'orange',
                   's': 'green',
                   'h': 'blue',
                   'i': 'yellow',
                   'm': 'red'}

    if ax is None:
        ax = plt.gca()

    if offset is None:
        offset = (0, 0)
    elif offset is True:
        offset = (ax.get_xlim()[1], ax.get_ylim()[1])
    else:
        pass

    vrna_coords = RNA.get_xy_coordinates(bp_string)
    # TODO Add option to rotate the plot
    for i, _ in enumerate(bp_string):
        coord = (offset[0] + vrna_coords.get(i).X,
                 offset[1] + vrna_coords.get(i).Y)
        coords.append(coord)
    coords = np.array(coords)
    # First plot backbone
    bkwargs = {"color":"black", "zorder":0}
    bkwargs.update(backbone_kwargs)
    ax.plot(coords[:,0], coords[:,1], **bkwargs)
    # Now plot basepairs
    basepairs = []
    for s in cg.stem_iterator():
        for p1, p2 in cg.stem_bp_iterator(s):
            basepairs.append([coords[p1-1], coords[p2-1]])
    if basepairs:
        basepairs = np.array(basepairs)
        if color:
            c = "red"
        else:
            c = "black"
            # Do not allow changing color through basepair_kwargs if color is set to False
            basepair_kwargs["color"] = c
        bpkwargs = {"color":c, "zorder":0, "linewidth":3}
        bpkwargs.update(basepair_kwargs)
        ax.plot(basepairs[:,:,0].T, basepairs[:,:,1].T, **bpkwargs)
    # Now plot circles
    for i, coord in enumerate(coords):
        if color:
            c = el_to_color[el_string[i]]
            h,l,s = colorsys.rgb_to_hls(*mc.to_rgb(c))
            if lighten>0:
                l += (1-l)*min(1,lighten)
            else:
                l +=l*max(-1, lighten)
            if l>1 or l<0:
                print(l)
            c=colorsys.hls_to_rgb(h,l,s)
            circle = plt.Circle((coord[0], coord[1]),
                            color=c)
        else:
            circle = plt.Circle((coord[0], coord[1]),
                                edgecolor="black", facecolor="white")

        ax.add_artist(circle)
        if cg.seq:
            if "fontweight" not in text_kwargs:
                text_kwargs["fontweight"]="bold"
            ax.annotate(cg.seq[i+1],xy=coord, ha="center", va="center", **text_kwargs )

    all_coords=list(coords)
    ntnum_kwargs = {"color":"gray"}
    ntnum_kwargs.update(text_kwargs)
    for nt in range(10, cg.seq_length, 10):
        # We try different angles
        annot_pos = _find_annot_pos_on_circle(nt, all_coords, cg)
        if annot_pos is not None:
            ax.annotate(str(nt), xy=coords[nt-1], xytext=annot_pos,
                        arrowprops={"width":1, "headwidth":1, "color":"gray"},
                        ha="center", va="center", zorder=0, **ntnum_kwargs)
            all_coords.append(annot_pos)

    _annotate_rna_plot(ax, cg, all_coords, annotations, text_kwargs)
    datalim = ((min(list(coords[:, 0]) + [ax.get_xlim()[0]]),
                min(list(coords[:, 1]) + [ax.get_ylim()[0]])),
               (max(list(coords[:, 0]) + [ax.get_xlim()[1]]),
                max(list(coords[:, 1]) + [ax.get_ylim()[1]])))

    '''
    min_coord = min(datalim[0][0], datalim[0][1])
    max_coord = max(datalim[1][0], datalim[1][1])
    datalim = ((min_coord, min_coord), (max_coord, max_coord))

    print "min_coord:", min_coord
    print "max_coord:", max_coord
    print "datalime:", datalim
    '''

    width = datalim[1][0] - datalim[0][0]
    height = datalim[1][1] - datalim[0][1]

    #ax.set_aspect(width / height)
    ax.set_aspect('equal', 'datalim')
    ax.update_datalim(datalim)
    ax.autoscale_view()
    ax.set_axis_off()

    return (ax, coords)

def _find_annot_pos_on_circle(nt, coords, cg):
    for i in range(5):
        for sign in [-1,1]:
            a = np.pi/4*i*sign
            if cg.get_elem(nt)[0]=="s":
                bp = cg.pairing_partner(nt)
                anchor = coords[bp-1]
            else:
                anchor =np.mean([ coords[nt-2], coords[nt]], axis=0)
            vec = coords[nt-1]-anchor
            vec=vec/ftuv.magnitude(vec)
            rotated_vec =  np.array([vec[0]*math.cos(a)-vec[1]*math.sin(a),
                                     vec[0]*math.sin(a)+vec[1]*math.cos(a)])
            annot_pos = coords[nt-1]+rotated_vec*18
            if _clashfree_annot_pos(annot_pos, coords):
                log.debug("Annot pos on c is %s",annot_pos)
                return annot_pos
    return None

def plot_pdb(filename, ax=None):
    """
    Plot the secondary structure of an RNA in a PDB file using the
    Graph Layout from the ViennaRNA package and indicate long-range
    interations and RNA-protein interactions.

    Interchain RNA-RNA interactions are shown as red lines.
    Proteins are shown as transparent gray circles with lines
    indicating the interacting residues. The circle radius corresponds
    to the number of interacting nucleotides.


    :param structure: A Bio.PDB.Structure
    :param ax: Optional. An matplotlib axis object
    :return: An Axes object (ax)
    """

    structure = Bio.PDB.PDBParser().get_structure('blah', filename)
    model = list(structure)[0]
    chain_coords = {}
    cgs = {}

    import collections as col

    # store a list of RNA nucleotides that each protein interacts with
    protein_interactions = col.defaultdict(set)
    protein_circles = []

    for chain in model:
        # iterate over RNAs
        if ftup.contains_rna(chain):
            # convert to cg and store so that we can convert pdb nucleotide ids to
            # secondary structure indexes later
            cg, = ftmc.CoarseGrainRNA.from_pdb(filename, load_chains=chain.id)
            cgs[chain.id] = cg

            # plot the structure and store the coordinates
            (ax, coords) = plot_rna(cg, offset=True, ax=ax)
            chain_coords[chain.id] = coords

    for (a1, a2) in ftup.interchain_contacts(structure):
        # iterate over all the interactions in order to find out which
        # nucleotides this protein interacts with
        chain1 = a1.parent.parent
        chain2 = a2.parent.parent

        if ftup.is_protein(chain1) and ftup.contains_rna(chain2):
            # collect all the RNA nucleotides that a protein interacts with
            sid = cgs[chain2.id].seq.to_integer(fgr.RESID(chain2.id, a2.parent.id))-1
            protein_interactions[chain1.id].add((chain2.id, sid))

        if ftup.contains_rna(chain1) and ftup.contains_rna(chain2):
            try:
                sid1 = cgs[chain1.id].seq.to_integer(fgr.RESID(chain1.id, a1.parent.id))-1
                sid2 = cgs[chain2.id].seq.to_integer(fgr.RESID(chain2.id, a2.parent.id))-1
            except ValueError:
                continue

            coord1 = chain_coords[chain1.id][sid1]
            coord2 = chain_coords[chain2.id][sid2]

            ax.plot([coord1[0], coord2[0]], [coord1[1], coord2[1]],
                    'k-', alpha=0.5)

    for chain in model:
        # draw each protein and the links that it has to other nucleotides
        if ftup.is_protein(chain):
            # the protein will be positioned at the centroid of the nucleotides
            # that it interacts with
            interacting_coords = [np.array(chain_coords[chain_id][nuc_num])
                                  for (chain_id, nuc_num) in protein_interactions[chain.id]]

            centroid = np.sum(interacting_coords, axis=0) / \
                len(interacting_coords)

            # the size of the circle representing it will be proportional to its
            # length (in nucleotides)
            radius = 2 * math.sqrt(len(chain.get_list()))
            protein_circles += [[centroid[0], centroid[1], radius]]

            # draw all of the interactions as lines
            for coord in interacting_coords:
                ax.plot([coord[0], centroid[0]], [
                        coord[1], centroid[1]], 'k-', alpha=0.5)

    protein_circles = np.array(protein_circles)
    if len(protein_circles) > 0:
        circles(protein_circles[:, 0], protein_circles[:, 1],
                protein_circles[:, 2], 'grey', alpha=0.5)

    # plt.axis('off')

    return ax



from __future__ import division
from builtins import range

import math
import warnings
import sys
import forgi.threedee.utilities.vector as ftuv
import itertools as it
import logging

import numpy as np

log = logging.getLogger(__name__)
"""
This module contains functions for characterising a single point-cloud.
"""


def radius_of_gyration(coords):
    '''
    Calculate the radius of gyration, given a set of coordinates.
    '''
    centroid = sum(coords) / float(len(coords))
    diff_vecs = coords - centroid
    # cud.pv('diff_vecs')
    sums = np.sum(diff_vecs * diff_vecs, axis=1)
    # cud.pv('sums')
    total = sum(sums)
    total /= len(coords)
    rmsd = math.sqrt(total)

    return rmsd


def gyration_tensor(coords, diagonalize=True):
    '''
    Calculate the gyration tensor, given a list of 3D coordinates.

    The gyration tensor is defiend as in doi:10.1063/1.4788616, eq. 4

    :param diagonalize: Diagonalize the tensor to diag(lambda1, lambda2, lambda3)
    '''
    if len(coords) == 0:
        log.warning("Cannot calculate gyration tensor: coords are empty, returning 'nan'")
        return np.zeros((3, 3)) * float("nan")
    if len(coords[0]) != 3:
        raise ValueError("Coordinates for Gyration Tensor must be in 3D space")
    centroid = sum(coords) / float(len(coords))
    diff_vecs = coords - centroid
    tensor = np.zeros((3, 3))
    tensor[0, 0] = sum(diff_vecs[:, 0]**2)
    tensor[1, 1] = sum(diff_vecs[:, 1]**2)
    tensor[2, 2] = sum(diff_vecs[:, 2]**2)
    tensor[0, 1] = tensor[1, 0] = sum(diff_vecs[:, 0] * diff_vecs[:, 1])
    tensor[0, 2] = tensor[2, 0] = sum(diff_vecs[:, 0] * diff_vecs[:, 2])
    tensor[1, 2] = tensor[2, 1] = sum(diff_vecs[:, 1] * diff_vecs[:, 2])
    if not diagonalize:
        return tensor
    tensor /= len(coords)
    eigenvalues = np.linalg.eigvals(tensor)
    assert len(eigenvalues) == 3
    tensor = np.zeros((3, 3))
    for i, v in enumerate(sorted(eigenvalues, reverse=True)):
        tensor[i, i] = v
    return tensor


def anisotropy(coords):
    """
    Calculate the anisotropy of a list of points in 3D space

    See for example doi:10.1063/1.4788616
    """
    g_tensor = gyration_tensor(coords)
    eigVs = [g_tensor[i, i] for i in range(3)]
    pp = 0
    for eV1, eV2 in it.combinations(eigVs, 2):
        pp += eV1 * eV2
    return 1 - 3 * (pp) / (sum(eigVs)**2)


def asphericity(coords):
    """
    Calculate the asphericity of a list of points in 3D space

    See for example doi:10.1063/1.4788616
    """
    g_tensor = gyration_tensor(coords)
    return g_tensor[0, 0] - (g_tensor[1, 1] + g_tensor[2, 2]) / 2.



from .graph import bulgegraph as fgb
from .threedee.model import stats as ftms
from .threedee.utilities import graph_pdb as ftug
from .threedee.utilities import pdb as ftup
from .threedee.utilities import vector as ftuv
"""
This is used for vulture (dead code detection tool) 
to mark library-functions as useful, even if they
are not used within the forgi project.
"""

fgb.print_brackets
bg = fgb.BulgeGraph()
bg.interior_loop_iterator()
bg.is_single_stranded()

s = ftms.ContinuousAngleStats()
ftug.spos_to_pos
ftup.output_chain.HSelect.accept_atom

ftuv.null_array
ftuv.tau
ftuv.is_almost_colinear([1], [2])
ftuv.get_standard_basis(3)
ftuv.vec_dot([0, 2, 2], [3, 4, 5])



from __future__ import absolute_import, unicode_literals, division
from __future__ import print_function
from builtins import (ascii, bytes, chr, dict, filter, hex, input, # pylint: disable=W0611
                      map, next, oct, pow, range, round, # pylint: disable=W0611
                      str, super, zip) # pylint: disable=W0611

try:
    from collections.abc import Sequence as SequenceABC
except ImportError:
    from collections import Sequence as SequenceABC

import logging
from collections import defaultdict, Counter
from string import ascii_lowercase, ascii_uppercase
from functools import partial
import inspect

from logging_exceptions import log_to_exception
from . import residue as fgr
log = logging.getLogger(__name__)

VALID_CHAINIDS = ascii_uppercase + ascii_lowercase


class MissingResidue(object):
    def __init__(self, resid, res_name):
        """
        :param resid: A fgr.RESID instance of a string.
        :param resname: A string. Probably one of "AUGC".
        """
        if not isinstance(resid, fgr.RESID):
            log.debug("Type of %s is %s", resid, type(resid).__name__)
            resid = fgr.resid_from_str(resid)
        self.resid = resid
        self.res_name = res_name

    def to_bg_string(self):
        return "missing {} {}".format(fgr.resid_to_str(self.resid), self.res_name)

    @classmethod
    def from_bg_fields(cls, parts):
        """
        Used during loading of the bg_file.

        :param parts: A list of strings. If the first is not "missing",
                      None is returned
        :returns: Either a MissingResidue instance or None
        """
        if parts[0] != "missing":
            return None
        return cls(fgr.resid_from_str(parts[1]), parts[2])


class _Smallest(object):
    """
    Smaller than everything, regardless of type
    """

    def __lt__(self, other):
        return True


class _Biggest(object):
    """
    Bigger than everything, regardless of type.
    """

    def __gt__(self, other):
        return True


def to_0_based(key):
    log.debug("Converting key %s", key)
    if isinstance(key, slice):
        # Step
        step = key.step
        if step is None:
            step = 1
        if abs(step) != 1:
            raise IndexError("Slices with step != 1 are not supported!")
        # Start
        start = key.start
        if start is None:
            start = None
        elif start > 0:
            start = start - 1
        else:
            raise IndexError("Slices with start <1 are currently not defined.")
        # Stop
        if step < 0:
            if key.stop is None:
                stop = None
            elif key.stop < -1:
                raise IndexError(
                    "Slices with negative stop and negative step are currently not defined.")
            else:
                stop = key.stop - 2
        else:
            stop = key.stop
        key = slice(start, stop, step)
    elif isinstance(key, int):
        if key == 0:
            raise IndexError("Index 0 not allowed in 1-based indexing")
        elif key > 0:
            key = key - 1
    else:
        raise TypeError("Invalid index type {}".format(type(key).__name__))
    log.debug("Returning key %s", key)
    return key


def _insert_breakpoints_simple(seq, breakpoints, start=0, reverse=False):
    """
    :param seq: A sequence without any missing residues
    :param breakpoints: 0-based break points
    :param start: The coordinate "on the forward strand",
                 i.e. the lower number of start and stop.
    """
    log.debug("Inserting breakpoints %s into %s", seq, breakpoints)
    breakpoints = sorted(breakpoints)
    if not breakpoints:
        return seq
    if start is None:
        start = 0
    out = []
    if reverse:
        seq = seq[::-1]
    log.debug("Inserting breakpoints into %s, start=%s, "
              "reversed=%s", seq, start, reverse)
    oldbp = 0
    for bp in breakpoints:
        log.debug("BP=%s", bp)
        bp -= start
        log.debug("bp-=start-->>>%s", bp)
        out.append(seq[oldbp:bp + 1])
        oldbp = bp + 1
    out.append(seq[bp + 1:]) # pylint: disable=undefined-loop-variable
    out = [o for o in out if o]
    seq = "&".join(out)
    if reverse:
        seq = seq[::-1]
    log.debug("seq with break points: %s", seq)
    return seq


def _resid_key(x):
    if x.resid[2] is None:
        return (x.resid[1], " ")
    else:
        # Convert to newstring on python2 See: https://github.com/PythonCharmers/python-future/issues/110
        return (x.resid[1], str(x.resid[2]))


def _sorted_missing_residues(list_of_dicts):
    """
    :param list_of_dicts: A list of dicts, as in my PR to biopython
                          or with the two keys "RESID" and "res_name",
    """
    chain_to_residues = defaultdict(list)
    resid_to_nucleotide = {}
    for res_dict in list_of_dicts:
        log.debug("Processing %s", res_dict)
        if isinstance(res_dict, MissingResidue):
            resid = res_dict.resid
            chain = res_dict.resid.chain
            res_name = res_dict.res_name
        elif "RESID" in res_dict:
            resid = res_dict["RESID"]
            if not isinstance(resid, fgr.RESID):
                resid = fgr.resid_from_str(resid)
            chain = resid.chain
            res_name = res_dict["res_name"]
        else:
            if res_dict["model"] not in [None, 1, "A"]:
                log.info("Invalid missing residue %s", res_dict)
                continue
            if res_dict["insertion"] is None:
                insertion = " "
            else:
                insertion = res_dict["insertion"]
            chain = res_dict["chain"]
            resid = fgr.RESID(chain, (' ', res_dict["ssseq"], insertion))
            res_name = res_dict["res_name"]
        chain_to_residues[chain].append(resid)
        resid_to_nucleotide[resid] = res_name
    for reslist in chain_to_residues.values():
        log.debug("Sorting %s", reslist)
        reslist.sort(key=_resid_key)
    return chain_to_residues, resid_to_nucleotide


class _IndexHelper(object):
    # pylint: disable=protected-access
    @property
    def flag(self):
        raise NotImplementedError("Must be set in subclass")

    def __init__(self, parent):
        self.parent = parent

    def __getitem__(self, key):
        log.debug("%s.__getitem__ called. %s=True",
                  type(self).__name__, self.flag)
        return self.parent._getitem(key, **{self.flag: True})

    def _getitem(self, key, include_missing=False, show_modifications=False):
        kwargs = {"include_missing": include_missing,
                  "show_modifications": show_modifications}
        kwargs.update({self.flag: True})
        log.debug("%s._getitem called. flags: %s", type(self).__name__, kwargs)
        return self.parent._getitem(key, **kwargs)

    def __getattr__(self, attr):
        """
        If attr is a method that supports **kwargs or has an attribute with
        the same name as the value of self.flag, then we set this
        attributre to True
        """
        log.debug("Getattr called for %s", attr)
        f = getattr(self.parent, attr)
        if callable(f):
            try:
                argspec = inspect.getfullargspec(f)
            except AttributeError:
                argspec = inspect.getargspec(f)
            log.debug("For function %s: args are %s", f, argspec.args)
            if (getattr(argspec, "varkw", getattr(argspec, "keywords", None)) 
                    or self.flag in argspec.args):
                log.debug("setting flag %s", self.flag)
                kwargs = {self.flag: True}
                f = partial(f, **kwargs)
        return f
    def __str__(self):
        return str(self[:])

class _WMIndexer(_IndexHelper):
    # pylint: disable=protected-access

    flag = "include_missing"

    def __len__(self):
        return len(self.parent) + len(self.parent._missing_nts)

    def update_dotbracket(self, struct):
        """
        Given a dotbracket_string of the same length as the sequence,
        update it with '-' for missing residues and '&' for breakpoints.

        :returns: A string
        """
        return self.parent._missing_into_db(struct)

    def export_missing(self):
        """
        Return the missing residues in a format which can be
        passed to the constructor of a new Sequence object.
        """
        return self.parent._export_missing()

    @property
    def with_modifications(self):
        return _MODIndexer(self)

    def define_length(self, d):
        val = 0
        for i in range(0, len(d), 2):
            if d[i+1]>=d[i]:
                val += sum(1 for _ in self.iter_resids(
                    self.to_resid(d[i]), self.to_resid(d[i + 1])))
                log.debug("Define length of %s with missing incremented to %s", d, val)
        log.debug("Define length of %s with missing is finally %s", d, val)
        return val


class _MODIndexer(_IndexHelper):
    flag = "show_modifications"

    @property
    def with_missing(self):
        return _WMIndexer(self)


class SeqidList(SequenceABC):
    def __init__(self, arg):
        self._list = list(arg)
        self._lookup = {resid: i for i, resid in enumerate(self._list)}
        if len(self._lookup) != len(self._list):
            # duplicate seqids
            c = Counter(self._list)
            for k, amount in c.most_common():
                if amount > 1:
                    log.error("Seq_id %s  occurs %s times!", k, amount)
                else:
                    break
            if c.most_common()[0][1] == 1:
                raise ValueError("len %{} ({}) != ({}) len {}".format(
                                            self._lookup, len(self._lookup),
                                            len(self._list), self._list))
            raise ValueError(
                "Duplicate Seq_id encountered: {}".format(c.most_common()[0][0]))

    def __getitem__(self, i):
        return self._list[i]

    def __len__(self):
        return len(self._list)

    def index(self, elem):
        try:
            return self._lookup[elem]
        except KeyError:
            raise ValueError("{} not in list".format(elem))

    def __ne__(self, other):
        if not isinstance(other, SeqidList):
            return NotImplemented
        return not self==other

    def __eq__(self, other):
        if isinstance(other, list):
            return self._list == other
        elif not isinstance(other, SeqidList):
            return NotImplemented
        return self._list == other._list

    def __hash__(self):
        return hash(self._list)

    def __repr__(self):
        return "SeqidList({})".format(repr(self._list))

class Sequence(object):
    """
    A class holding the RNA sequence.

    It supports 2 indexing conventions (see below), missing residues
    and modified nucleotids.

    The two indexing conventions are:

    *  1-based indexing into the sequence for which structure information is available
       This uses indices of type integer.
       Integer based indexing can not address missing residues!
    *  Indexing using PDB-style numbers (They can start at any positive or negative number and
       may contain insertion codes or gaps.)
       This uses indices of type fgr.RESID
       PDB-Style indices may address missing residues (Reidues for which only sequence
       but no structure information is present).
    """

    def __init__(self, seq, seqids, missing_residues=None, modifications=None):
        """
        :param seq: A string
        :param missing_residues: A list of dictionaries with the following keys:
                "res_name", "chain", "ssseq", "insertion"
        """
        log.debug("Sequence initialized with %s, %s, %s", seq, list(
            map(fgr.resid_to_str, seqids)), missing_residues)
        # Uses 0-based indexing
        i = 0
        self._breaks_after = []
        remaining = seq
        while remaining:
            c, remaining = remaining[0], remaining[1:]
            if c == '&':
                self._breaks_after.append(i - 1)
            else:
                i += 1
        log.debug("Break-points for seq %s are: %s", seq, self._breaks_after)
        self._seq = seq.replace('&', '')
        self._seqids = SeqidList(seqids)
        self._missing_residues = defaultdict(list)
        self._missing_nts = {}
        if missing_residues:
            self._set_missing_residues(missing_residues)  # A dict seq_id:nt
        # resid : modified res_name
        self._modifications = {}
        if modifications:
            self._modifications.update(modifications)

    def _set_missing_residues(self, missing_residues):
        mr, mnts = _sorted_missing_residues(missing_residues)
        log.debug("Setting missing residues to: %s, %s", mr, mnts)
        self._missing_residues = mr
        self._missing_nts = mnts  # A dict seq_id:nt

    def __str__(self):
        return self[:]

    def __repr__(self):
        try:
            qname = type(self).__qualname__
        except AttributeError:
            qname = type(self).__name__
        return "<{} object with ._seq={}>".format(qname, str(self))

    def __bool__(self):
        return bool(self._seq)
    __nonzero__ = __bool__

    @property
    def backbone_breaks_after(self):
        """
        Breakpoints as 1-based indices
        """
        return [bp + 1 for bp in self._breaks_after]

    def is_valid(self):
        wrong_chars = set(self._seq) - set("AUGCaugc")
        if wrong_chars:
            log.info("Illegal characters are {}".format(wrong_chars))
            return False
        return True

    def __getitem__(self, key):
        """
        Never return missing residues!
        """
        return self._getitem(key, False)

    def __len__(self):
        return len(self._seq)

    def __hash__(self):
        raise AttributeError()

    def __eq__(self, other):
        log.debug("{} =?= {}".format(repr(self), repr(other)))
        log.debug("type of other=%s, type of string literal in this module is %s", type(
            other).__name__, type("").__name__)
        if isinstance(other, type(self)):
            if self._seq != other._seq:
                log.debug("Sequence different!")
                return False
            if self._breaks_after != other._breaks_after:
                log.debug("Breakpoints different!")
                return False
            if self._seqids != other._seqids:
                log.debug("seq_ids different: %s != %s",
                          self._seqids, other._seqids)
                return False
            if self._missing_nts != other._missing_nts:
                if log.isEnabledFor(logging.DEBUG):
                    for key, val in self._missing_nts.items():
                        if key in other._missing_nts:
                            if self._missing_nts[key] != other._missing_nts[key]:
                                log.debug("nt changed for %s from %s to %s", key,
                                          self._missing_nts[key], other._missing_nts[key])
                        else:
                            log.debug("%s missing in other", key)
                    for key in other._missing_nts:
                        if key not in self._missing_nts:
                            log.debug("%s extra in other", key)
                return False
            log.debug("They are equal")
            return True
        return str(self) == other

    def _getitem(self, key, include_missing=False, show_modifications=False):
        log.debug("_getitem called for %s, include_missing=%s, show_modifications=%s",
                  key, include_missing, show_modifications)
        if isinstance(key, int):
            key = to_0_based(key)
            if show_modifications and self._seqids[key] in self._modifications:
                return self._modifications[self._seqids[key]]
            else:
                return self._seq[key]
        elif isinstance(key, fgr.RESID):
            try:
                i = self._seqids.index(key)
            except ValueError:
                if key in self._missing_nts:
                    nt = self._missing_nts[key]
                    if include_missing == False:
                        raise IndexError("No structure available for nucleotide '{}'."
                                         "For look-up including missing residues, use"
                                         "`.with_missing[key]`".format(key))
                    if show_modifications and key in self._modifications:
                        return self._modifications[key]
                    return nt
                error = IndexError(
                    "Nucleotide {} is not part of this RNA".format(key))
                with log_to_exception(log, error):
                    log.error("self._missing_nts = %s", self._missing_nts)
                raise error
            else:
                if show_modifications and key in self._modifications:
                    return self._modifications[key]
                return self._seq[i]
        elif isinstance(key, slice):
            return self._getslice(key, include_missing, show_modifications)
        else:
            raise TypeError("Wrong index type: {}".format(type(key).__name__))

    def _getslice(self, key, include_missing=False, show_modifications=False):
        """
        Do type checks and delegate to self._resid_slice or self._integer_slice
        """
        if key.step is not None and key.step not in [1, -1]:
            raise IndexError("The step in the slice has to be 1 or -1")
        if not isinstance(key.start, (type(None), int, fgr.RESID)):
            raise TypeError(
                "Slice-start must be either an integer, a RESID or None")
        if not isinstance(key.stop, (type(None), int, fgr.RESID)):
            raise TypeError(
                "Slice-stop must be either an integer, a RESID or None")
        if isinstance(key.start, fgr.RESID) or (key.start is None and isinstance(key.stop, fgr.RESID)):
            if not isinstance(key.stop, (fgr.RESID, type(None))):
                raise TypeError(
                    "Mixing integers and RESID for slicing is not allowed")
            return self._resid_slice(key, include_missing, show_modifications)
        return self._integer_slice(to_0_based(key), include_missing, show_modifications)

    def _integer_slice(self, key, include_missing=False, show_modifications=False):
        """
        This provides an implementation for include_missing=False
        and delegates to resid_slice otherwise.

        :param key: Already 0-based, no None in slice
        """
        if ((include_missing == False or not self._missing_nts) and
                (show_modifications == False or not self._modifications)):
            seq = self._seq[key]
            if key.step == -1:
                start = key.stop
                if start is not None:
                    start += 1
            else:
                start = key.start
            return _insert_breakpoints_simple(seq, self._breaks_after, start, key.step == -1)
        if key.start is None:
            startRES = None
        else:
            startRES = self._seqids[key.start]
        if key.stop is None:
            stopRES = None
        else:
            if key.step == -1:
                stopRES = self._seqids[key.stop + 1]
            else:
                if key.stop > 0:
                    stopRES = self._seqids[key.stop - 1]
                else:
                    stopRES = self._seqids[key.stop]
        key2 = slice(startRES, stopRES, key.step)
        log.debug("Converted integer_slice %s to resid-slice %s", key, key2)
        return self._resid_slice(key2, include_missing, show_modifications)

    def _resid_slice_to_int_slice(self, key):
        if key.step == -1:
            default_start = len(self._seqids)
            default_stop = None
        else:
            default_start = 0
            default_stop = len(self._seqids)
        start_i, _ = self._resid_to_index(key.start, default_start, False)
        stop_i, _ = self._resid_to_index(key.stop, default_stop, False)
        if key.step == -1:
            if stop_i is not None:
                stop_i -= 1
        else:
            stop_i += 1
        key = slice(start_i, stop_i, key.step)
        log.debug("Resid-slice mapped to integer-slice {}".format(key))
        return self._integer_slice(key, False)

    def _resid_to_index(self, resid, default, default_on_error=False):
        """
        :returns: a tuple index, error_raised
        """
        if resid is None:
            return default, False
        try:
            return self._seqids.index(resid), False
        except ValueError:
            if default_on_error:
                return default, True
            else:
                raise IndexError("RESID {} is not present or no structure"
                                 "is available for it.".format(resid))

    def _resid_slice(self, key, include_missing, show_modifications):
        """
        This provides an implementation for inclue_missing=True
        and delegates to integer_slice otherwise.

        :param key: A validated resid slice.
        """
        if include_missing or show_modifications:
            return self._resid_slice_main(key, include_missing, show_modifications)
        else:
            return self._resid_slice_to_int_slice(key)

    def _resid_slice_main(self, key, with_missing, show_modifications):
        if key.step == -1:
            start, stop = key.stop, key.start
        else:
            start, stop = key.start, key.stop
        seqids = list(self.iter_resids(start, stop, with_missing))
        if key.step == -1:
            seqids.reverse()
        seqs = [[]]
        seq = seqs[-1]  # a reference
        for seqid in seqids:
            if seqid == "&":
                if show_modifications:
                    seqs.append([])
                    seq = seqs[-1]
                else:
                    seq.append("&")
            else:
                if show_modifications and seqid in self._modifications:
                    seq.append(self._modifications[seqid])
                else:
                    try:
                        seq.append(self._seq[self._seqids.index(seqid)])
                    except ValueError:
                        seq.append(self._missing_nts[seqid])
        if not show_modifications:
            assert seqs == [seq]
            seqs = "".join(seq)
        return seqs

    def _missing_into_db(self, dotbracket):
        db = ""
        check_i = 0
        for seqid in self.iter_resids(None, None, True):
            if seqid == "&":
                db += "&"
            else:
                try:
                    i = self._seqids.index(seqid)
                    assert i == check_i
                except ValueError:
                    db += "-"
                else:
                    db += dotbracket[i]
                    check_i += 1
        return "".join(db)

    def iter_resids(self, start, stop, include_missing):
        left_res = None
        start_i, start_is_missing = self._resid_to_index(start, 0, True)
        stop_i, stop_is_missing = self._resid_to_index(
            stop, len(self._seqids), True)
        if (start_is_missing or stop_is_missing) and not include_missing:
            raise IndexError(
                "Start or stop are missing residues, but indexing without missing residues was requested.")

        if not start_is_missing and start is not None:
            left_res = start

        # Flag used later if missing_residues == True
        found_start = not start_is_missing
        seq = ""
        if left_res is not None:
            yield left_res
            # seq+=self._seq[self._seqids.index(left_res)]
            #log.debug("Added first residue %s: %s", left_res, seq)
            start_i += 1
        log.debug("Iterating over sequence %s-%s", start_i, stop_i + 1)
        mr_keys = None
        for i in range(start_i, stop_i + 1):
            try:
                right_res = self._seqids[i]
            except IndexError:
                right_res = None
            # Some consistency checks:
            if left_res is not None and right_res is not None and left_res.chain != right_res.chain:
                if i - 1 not in self._breaks_after:
                    raise ValueError("Inconsist break-points discovered."
                                     "Missing breakpoint at position {} "
                                     "between chain {} and chain {}. "
                                     "Break points are {}".format(i - 1,
                                                                  left_res.chain, right_res.chain, self._breaks_after))
            if i - 1 in self._breaks_after:
                if left_res is not None and right_res is not None and left_res.chain == right_res.chain:
                    raise ValueError("Inconsist break-points discovered."
                                     "Breakpoint at position {} "
                                     "in the middle of chain {}".format(i - 1,
                                                                        left_res.chain))
            if include_missing:
                mr, mr_keys = self._missing_residues_between(
                    left_res, right_res)
                old_r = left_res
                for j, r in enumerate(mr_keys):
                    if not found_start and r == start:
                        log.debug("Found start for i=%d, %s", i, r)
                        found_start = True
                        old_r = None
                    log.debug("Now processing missing residue %s", r)
                    if found_start:
                        if old_r is not None and old_r.chain != r.chain:
                            yield "&"
                            log.debug("Breakpoint inserted!")
                        log.debug("Adding missing residue %s", mr[j])
                        yield r
                    log.debug("Comparing %s and stop=%s", r, stop)
                    if r == stop:
                        log.debug("Stop found! Seq=%s", seq)
                        return
                    old_r = r
            # If key.start and key.stop resp. are not modified residues,
            # start_i and stop_i are already accurrate, making
            # additional checks here unneccessary.
            if found_start and i < len(self._seq):
                log.debug("Adding regular residue %s", self._seq[i])
                if mr_keys:  # No missing residues
                    if r.chain != right_res.chain:
                        yield "&"
                        log.debug("Breakpoint inserted!")
                else:
                    if left_res is not None and right_res is not None and left_res.chain != right_res.chain:
                        yield "&"
                        log.debug("Breakpoint inserted!")
                yield right_res
            left_res = right_res
            log.debug("seq now is %s", seq)
        # If start or stop are not part of sequence,
        # make sure they were seen in missing residues
        if not found_start or stop_is_missing:
            raise IndexError("At least one of the residues {} and {} "
                             "is not part of this RNA".format(start, stop))
        return

    def _missing_residues_between(self, from_resid, to_resid):
        log.debug("Searching missing residues between %s  and %s",
                  from_resid, to_resid)
        if from_resid is not None and to_resid is not None:
            if from_resid.chain != to_resid.chain:
                left = self._missing_residues_between(from_resid, None)
                right = self._missing_residues_between(None, to_resid)
                return left[0] + right[0], left[1] + right[1]
        if from_resid is not None:
            from_key = _resid_key(from_resid)
            chain = from_resid.chain
        else:
            from_key = _Smallest()
        if to_resid is not None:
            to_key = _resid_key(to_resid)
            chain = to_resid.chain
        else:
            to_key = _Biggest()
        seq = ""
        res_ids = []
        for res1 in self._missing_residues[chain]:
            if _resid_key(res1) > from_key and _resid_key(res1) < to_key:
                seq += self._missing_nts[res1]
                res_ids.append(res1)
        return seq, res_ids

    def to_resid(self, i):
        return self._seqids[i - 1]

    def to_integer(self, resid):
        return self._seqids.index(resid) + 1

    def iter_modifications(self):
        """
        Iter over tuples seq_id, modified residue code
        """
        for k in sorted(self._modifications):
            yield k, self._modifications[k]

    @property
    def with_missing(self):
        """
        Allows slices with missing residues!
        """
        if self._seqids:
            return _WMIndexer(self)
        else:  # For performance reasons
            return self

    @property
    def with_modifications(self):
        if self._modifications:
            return _MODIndexer(self)
        else:
            return self

    def get_bg_str(self):
        """
        Used during bg-file creation
        """
        out = []
        out.append("seq {}".format(str(self)))
        out.append("seq_ids {}".format(
            " ".join(map(fgr.resid_to_str, self._seqids))))
        for resid, nt in self._missing_nts.items():
            out.append(MissingResidue(resid, nt).to_bg_string())
        for resid, label in self._modifications.items():
            out.append("modification {} {}".format(
                fgr.resid_to_str(resid), label))
        return "\n".join(out) + "\n"

    def define_length(self, d):
        val = 0
        for i in range(0, len(d), 2):
            val += d[i + 1] - d[i] + 1
        log.debug("Define length of %s without missing is %s", d, val)

        return val

    def _export_missing(self):
        """
        Implementation for Sequence().with_missing.export_missing
        """
        out = []
        for resid, nt in self._missing_nts:
            out.append({"RESID": resid, "res_name": nt})
        return out

    def __add__(self, other):
        #print("Sequence add called")
        if isinstance(other, Sequence):
            raise NotImplementedError(
                "This is not yet implemented (but really should be)")
        elif isinstance(other, str):
            return str(self) + other
        else:
            return NotImplemented

    def __radd__(self, other):
        #print("Sequence radd called, other is ", repr(other), type(other).__name__, str.__name__)
        #print(isinstance(other, str))
        if isinstance(other, Sequence):
            raise NotImplementedError(
                "This is not yet implemented (but really should be)")
        elif isinstance(other, str):
            #print("Other is str")
            return other + str(self)
        else:
            return NotImplemented


class SequenceLoader:
    """
    Load the sequence-related part during loading of bg-files
    """

    def __init__(self):
        self.mod = {}
        self.mr = []
        self.seq = None
        self.seq_ids = []

    def consume_fields(self, parts):
        """
        Read one splitted line from the forgi file.

        Returns True, if the line was used/ understood, False otherwise
        """
        mr = MissingResidue.from_bg_fields(parts)
        if mr is not None:
            self.mr.append(mr)
            return True
        elif parts[0] == "seq":
            if self.seq is not None:
                raise ValueError("More than one seq-line encountered.")
            self.seq = parts[1]
            return True
        elif parts[0] == "seq_ids":
            if self.seq_ids:
                raise ValueError("More than one seq-ids line encountered.")
            self.seq_ids = list(map(fgr.resid_from_str, parts[1:]))
            return True
        elif parts[0] == "modification":
            self.mod[fgr.resid_from_str(parts[1])] = " ".join(parts[2:])
            return True
        return False

    @property
    def sequence(self):
        if self.seq is None and not self.seq_ids:
            # The breakpoints are only persisted at the seq and seq_id level.
            raise ValueError("Parsing incomplete: No seq found.")
        elif self.seq is None:
            old_chain = None
            seq = ""
            for resid in self.seq_ids:
                if old_chain is not None and resid.chain != old_chain:
                    seq += "&"
                seq += "N"
                old_chain = resid.chain
            self.seq = seq
        elif not self.seq_ids:
            self.seq_ids = _seq_ids_from_seq_str(self.seq)
        return Sequence(self.seq, self.seq_ids, self.mr, self.mod)


def _seq_ids_from_seq_str(seq):
    """
    Get a list of seq_ids with the same length as the sequence,
    respecting cutpoints.

    We start with seq_id A:1. If a '&' is encountered in the sequence,
    a new chain-letter is used.

    :param seq: A string, optionally containing '&' characters.
    """
    seq_strs = seq.split('&')
    seq_ids = []
    for i, seq_str in enumerate(seq_strs):
        for j, s in enumerate(seq_str):
            seq_ids += [fgr.resid_from_str(
                "{}:{}".format(VALID_CHAINIDS[i], j + 1))]
    return seq_ids


def missing_to_normal(seq):
    """
    Return a Sequence object which has all missing residues moved to normal residues.

    :param seq: A forgi.graph.sequence.Sequence object
    """
    new_seq = seq.with_missing[:]
    new_seqids = list(seq.with_missing.iter_resids(None,None))
    return Sequence(new_seq, new_seqids, None, seq._modifications)
